from typing import Any, Optional, Tuple, Type

import numpy as np
import torch
from torch import nn

from .base import BaseNeuronsLayer
from .. import HeavisideSigmoidApprox, SpikeFunction, HeavisidePhiApprox
from ...dimension import SizeTypes
from ...transforms import to_tensor
from ...utils import format_pseudo_rn_seed


# @inherit_fields_docstring(fields=["Attributes"], bases=[BaseNeuronsLayer])
class LIFLayer(BaseNeuronsLayer):
    """
    LIF dynamics, inspired by :cite:t:`neftci_surrogate_2019` , :cite:t:`bellec_solution_2020` , models the synaptic
    potential and impulses of a neuron over time. The shape of this potential is not considered realistic
    :cite:t:`izhikevich_dynamical_2007` , but the time at which the potential exceeds the threshold is.
    This potential is found by the recurrent equation :eq:`lif_V` .

    .. math::
        \\begin{equation}
            V_j^{t+\\Delta t} = \\left(\\alpha V_j^t + \\sum_{i}^{N} W_{ij}^{\\text{rec}} z_i^t +
            \\sum_i^{N} W_{ij}^{\\text{in}} x_i^{t+\\Delta t}\\right) \\left(1 - z_j^t\\right)
        \\end{equation}
        :label: lif_V


    The variables of the equation :eq:`lif_V` are described by the following definitions:

        - :math:`N` is the number of neurons in the layer.
        - :math:`V_j^t` is the synaptic potential of the neuron :math:`j` at time :math:`t`.
        - :math:`\\Delta t` is the integration time step.
        - :math:`z_j^t` is the spike of the neuron :math:`j` at time :math:`t`.
        - :math:`\\alpha` is the decay constant of the potential over time (equation :eq:`lif_alpha` ).
        - :math:`W_{ij}^{\\text{rec}}` is the recurrent weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`W_{ij}^{\\text{in}}` is the input weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`x_i^{t}` is the input of the neuron :math:`i` at time :math:`t`.

    .. math::
        :label: lif_alpha

        \\begin{equation}
            \\alpha = e^{-\\frac{\\Delta t}{\\tau_m}}
        \\end{equation}



    with :math:`\\tau_m` being the decay time constant of the membrane potential which is generally 20 ms.

    The output of neuron :math:`j` at time :math:`t` denoted :math:`z_j^t` is defined by the equation :eq:`lif_z` .

    .. math::
        :label: lif_z

        z_j^t = H(V_j^t - V_{\\text{th}})

    where :math:`V_{\\text{th}}` denotes the activation threshold of the neuron and the function :math:`H(\\cdot)`
    is the Heaviside function defined as :math:`H(x) = 1` if :math:`x \\geq 0` and :math:`H(x) = 0` otherwise.

    :Attributes:
        - :attr:`forward_weights` (torch.nn.Parameter): The weights used to compute the output of the layer :math:`W_{ij}^{\\text{in}}` in equation :eq:`lif_V`.
        - :attr:`recurrent_weights` (torch.nn.Parameter): The weights used to compute the hidden state of the layer :math:`W_{ij}^{\\text{rec}}` in equation :eq:`lif_V`.
        - :attr:`dt` (float): The time step of the layer :math:`\\Delta t` in equation :eq:`lif_V`.
        - :attr:`use_rec_eye_mask` (bool): Whether to use the recurrent eye mask.
        - :attr:`rec_mask` (torch.Tensor): The recurrent eye mask.
        - :attr:`alpha` (torch.nn.Parameter): The decay constant of the potential over time. See equation :eq:`lif_alpha` .
        - :attr:`threshold` (torch.nn.Parameter): The activation threshold of the neuron.
        - :attr:`gamma` (torch.nn.Parameter): The gain of the neuron. The gain will increase the gradient of the neuron's output.

    """

    # @inherit_docstring(bases=BaseNeuronsLayer)
    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            use_recurrent_connection: bool = True,
            use_rec_eye_mask: bool = False,
            spike_func: Type[SpikeFunction] = HeavisideSigmoidApprox,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        """
        :keyword float tau_m: The decay time constant of the membrane potential which is generally 20 ms. See equation
            :eq:`lif_alpha` .
        :keyword float threshold: The activation threshold of the neuron.
        :keyword float gamma: The gain of the neuron. The gain will increase the gradient of the neuron's output.
        :keyword float spikes_regularization_factor: The regularization factor of the spikes.
        """
        self.spike_func = spike_func
        super(LIFLayer, self).__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=use_recurrent_connection,
            use_rec_eye_mask=use_rec_eye_mask,
            dt=dt,
            device=device,
            **kwargs
        )

        self.alpha = nn.Parameter(
            torch.tensor(np.exp(-dt / self.kwargs["tau_m"]), dtype=torch.float32, device=self.device),
            requires_grad=False
        )
        self.threshold = nn.Parameter(
            torch.tensor(self.kwargs["threshold"], dtype=torch.float32, device=self.device),
            requires_grad=False
        )
        self.gamma = nn.Parameter(
            torch.tensor(self.kwargs["gamma"], dtype=torch.float32, device=self.device),
            requires_grad=False
        )

    def _set_default_kwargs(self):
        self.kwargs.setdefault("tau_m", 10.0 * self.dt)
        self.kwargs.setdefault("threshold", 1.0)
        if issubclass(self.spike_func, HeavisideSigmoidApprox):
            self.kwargs.setdefault("gamma", 100.0)
        else:
            self.kwargs.setdefault("gamma", 1.0)
        self.kwargs.setdefault("spikes_regularization_factor", 0.0)

    # @inherit_docstring(bases=BaseNeuronsLayer)
    def initialize_weights_(self):
        super().initialize_weights_()
        if self.kwargs.get("forward_weights", None) is not None:
            self.forward_weights.data = to_tensor(self.kwargs["forward_weights"]).to(self.device)
        else:
            torch.nn.init.xavier_normal_(self.forward_weights)

        if self.kwargs.get("recurrent_weights", None) is not None and self.use_recurrent_connection:
            self.recurrent_weights.data = to_tensor(self.kwargs["recurrent_weights"]).to(self.device)
        elif self.use_recurrent_connection:
            torch.nn.init.xavier_normal_(self.recurrent_weights)

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        """
        Create an empty state in the following form:
            ([membrane potential of shape (batch_size, self.output_size)],
            [spikes of shape (batch_size, self.output_size)])

        :param batch_size: The size of the current batch.
        :return: The current state.
        """
        kwargs.setdefault("n_hh", 2)
        return super(LIFLayer, self).create_empty_state(batch_size=batch_size, **kwargs)

    def update_regularization_loss(self, state: Optional[Any] = None, *args, **kwargs) -> torch.Tensor:
        """
        Update the regularization loss for this layer. Each update call increments the regularization loss so at the end
        the regularization loss will be the sum of all calls to this function.

        :param state: The current state of the layer.
        :return: The updated regularization loss.
        """
        next_V, next_Z = state
        self._regularization_loss += self.kwargs["spikes_regularization_factor"] * torch.sum(next_Z)
        # self._regularization_loss += 2e-6*torch.mean(torch.sum(next_Z, dim=-1)**2)
        return self._regularization_loss

    # @inherit_docstring(bases=BaseNeuronsLayer)
    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        assert inputs.ndim == 2
        batch_size, nb_features = inputs.shape
        V, Z = self._init_forward_state(state, batch_size, inputs=inputs)
        input_current = torch.matmul(inputs, self.forward_weights)
        if self.use_recurrent_connection:
            rec_current = torch.matmul(Z, torch.mul(self.recurrent_weights, self.rec_mask))
        else:
            rec_current = 0.0
        next_V = (self.alpha * V + input_current + rec_current) * (1.0 - Z.detach())
        next_Z = self.spike_func.apply(next_V, self.threshold, self.gamma)
        return next_Z, (next_V, next_Z)


# @inherit_fields_docstring(fields=["Attributes"], bases=[BaseNeuronsLayer])
class SpyLIFLayer(BaseNeuronsLayer):
    """
    The SpyLIF dynamics is a more complex variant of the LIF dynamics (class :class:`LIFLayer`) allowing it to have a
    greater power of expression. This variant is also inspired by Neftci :cite:t:`neftci_surrogate_2019` and also
    contains  two differential equations like the SpyLI dynamics :class:`SpyLI`. The equation :eq:`SpyLIF_I` presents
    the synaptic current update equation with euler integration while the equation :eq:`SpyLIF_V` presents the
    synaptic potential update.

    .. math::
        :label: SpyLIF_I

        \\begin{equation}
            I_{\\text{syn}, j}^{t+\\Delta t} = \\alpha I_{\text{syn}, j}^{t} + \\sum_{i}^{N} W_{ij}^{\\text{rec}} z_i^t
            + \\sum_i^{N} W_{ij}^{\\text{in}} x_i^{t+\\Delta t}
        \\end{equation}


    .. math::
        :label: SpyLIF_V

        \\begin{equation}
            V_j^{t+\\Delta t} = \\left(\\beta V_j^t + I_{\\text{syn}, j}^{t+\\Delta t}\\right) \\left(1 - z_j^t\\right)
        \\end{equation}


    .. math::
        :label: spylif_alpha

        \\begin{equation}
            \\alpha = e^{-\\frac{\\Delta t}{\\tau_{\\text{syn}}}}
        \\end{equation}

    with :math:`\\tau_{\\text{syn}}` being the decay time constant of the synaptic current.

    .. math::
        :label: spylif_beta

        \\begin{equation}
            \\beta = e^{-\\frac{\\Delta t}{\\tau_{\\text{mem}}}}
        \\end{equation}

    with :math:`\\tau_{\\text{syn}}` being the decay time constant of the synaptic current.

    The output of neuron :math:`j` at time :math:`t` denoted :math:`z_j^t` is defined by the equation :eq:`spylif_z` .

    .. math::
        :label: spylif_z

        z_j^t = H(V_j^t - V_{\\text{th}})

    where :math:`V_{\\text{th}}` denotes the activation threshold of the neuron and the function :math:`H(\\cdot)`
    is the Heaviside function defined as :math:`H(x) = 1` if :math:`x \\geq 0` and :math:`H(x) = 0` otherwise.

    SpyTorch library: https://github.com/surrogate-gradient-learning/spytorch.

    The variables of the equations :eq:`SpyLIF_I` and :eq:`SpyLIF_V` are described by the following definitions:

        - :math:`N` is the number of neurons in the layer.
        - :math:`I_{\\text{syn}, j}^{t}` is the synaptic current of neuron :math:`j` at time :math:`t`.
        - :math:`V_j^t` is the synaptic potential of the neuron :math:`j` at time :math:`t`.
        - :math:`\\Delta t` is the integration time step.
        - :math:`z_j^t` is the spike of the neuron :math:`j` at time :math:`t`.
        - :math:`\\alpha` is the decay constant of the synaptic current over time (equation :eq:`spylif_alpha`).
        - :math:`\\beta` is the decay constant of the membrane potential over time (equation :eq:`spylif_beta`).
        - :math:`W_{ij}^{\\text{rec}}` is the recurrent weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`W_{ij}^{\\text{in}}` is the input weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`x_i^{t}` is the input of the neuron :math:`i` at time :math:`t`.

    :Attributes:
        - :attr:`alpha` (torch.nn.Parameter): Decay constant of the synaptic current over time (equation :eq:`spylif_alpha`).
        - :attr:`beta` (torch.nn.Parameter): Decay constant of the membrane potential over time (equation :eq:`spylif_beta`).
        - :attr:`threshold` (torch.nn.Parameter): Activation threshold of the neuron (:math:`V_{\\text{th}}`).
        - :attr:`gamma` (torch.nn.Parameter): Slope of the Heaviside function (:math:`\\gamma`).
    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            use_recurrent_connection: bool = True,
            use_rec_eye_mask: bool = False,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        """
        Constructor for the SpyLIF layer.

        :param input_size: The size of the input.
        :type input_size: Optional[SizeTypes]
        :param output_size: The size of the output.
        :type output_size: Optional[SizeTypes]
        :param name: The name of the layer.
        :type name: Optional[str]
        :param use_recurrent_connection: Whether to use the recurrent connection.
        :type use_recurrent_connection: bool
        :param use_rec_eye_mask: Whether to use the recurrent eye mask.
        :type use_rec_eye_mask: bool
        :param spike_func: The spike function to use.
        :type spike_func: Callable[[torch.Tensor], torch.Tensor]
        :param learning_type: The learning type to use.
        :type learning_type: LearningType
        :param dt: Time step (Euler's discretisation).
        :type dt: float
        :param device: The device to use.
        :type device: Optional[torch.device]
        :param kwargs: The keyword arguments for the layer.

        :keyword float tau_syn: The synaptic time constant :math:`\\tau_{\\text{syn}}`. Default: 5.0 * dt.
        :keyword float tau_mem: The membrane time constant :math:`\\tau_{\\text{mem}}`. Default: 10.0 * dt.
        :keyword float threshold: The threshold potential :math:`V_{\\text{th}}`. Default: 1.0.
        :keyword float gamma: The multiplier of the derivative of the spike function :math:`\\gamma`. Default: 100.0.
        :keyword float spikes_regularization_factor: The regularization factor for the spikes. Higher this factor is,
            the more the network will tend to spike less. Default: 0.0.

        """
        self.spike_func = HeavisideSigmoidApprox
        super(SpyLIFLayer, self).__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=use_recurrent_connection,
            use_rec_eye_mask=use_rec_eye_mask,
            dt=dt,
            device=device,
            **kwargs
        )

        self.alpha = nn.Parameter(
            torch.tensor(np.exp(-dt / self.kwargs["tau_syn"]), dtype=torch.float32, device=self.device),
            requires_grad=False
        )
        self.beta = nn.Parameter(
            torch.tensor(np.exp(-dt / self.kwargs["tau_mem"]), dtype=torch.float32, device=self.device),
            requires_grad=False
        )
        self.threshold = nn.Parameter(
            torch.tensor(self.kwargs["threshold"], dtype=torch.float32, device=self.device),
            requires_grad=False
        )
        self.gamma = nn.Parameter(
            torch.tensor(self.kwargs["gamma"], dtype=torch.float32, device=self.device),
            requires_grad=False
        )
        self._regularization_l1 = torch.tensor(0.0, dtype=torch.float32, device=self.device)
        self._n_spike_per_neuron = torch.zeros(int(self.output_size), dtype=torch.float32, device=self.device)
        self._total_count = 0

    def _set_default_kwargs(self):
        self.kwargs.setdefault("tau_syn", 5.0 * self.dt)
        self.kwargs.setdefault("tau_mem", 10.0 * self.dt)
        self.kwargs.setdefault("threshold", 1.0)
        self.kwargs.setdefault("gamma", 100.0)
        self.kwargs.setdefault("spikes_regularization_factor", 0.0)
        self.kwargs.setdefault("hh_init", "zeros")

    def initialize_weights_(self):
        super().initialize_weights_()
        weight_scale = 0.2
        if self.kwargs.get("forward_weights", None) is not None:
            self.forward_weights.data = to_tensor(self.kwargs["forward_weights"]).to(self.device)
        else:
            torch.nn.init.normal_(self.forward_weights, mean=0.0, std=weight_scale / np.sqrt(int(self.input_size)))

        if self.kwargs.get("recurrent_weights", None) is not None and self.use_recurrent_connection:
            self.recurrent_weights.data = to_tensor(self.kwargs["recurrent_weights"]).to(self.device)
        elif self.use_recurrent_connection:
            torch.nn.init.normal_(self.recurrent_weights, mean=0.0, std=weight_scale / np.sqrt(int(self.output_size)))

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        """
        Create an empty state in the following form:
            ([membrane potential of shape (batch_size, self.output_size)],
            [synaptic current of shape (batch_size, self.output_size)],
            [spikes of shape (batch_size, self.output_size)])

        :param batch_size: The size of the current batch.
        :return: The current state.
        """
        kwargs.setdefault("n_hh", 3)
        thr = self.threshold.detach().cpu().item()
        if self.kwargs["hh_init"] == "random":
            V_mu, V_std = self.kwargs.get("hh_init_mu", thr / 2.0), self.kwargs.get("hh_init_std", 0.341 * thr)
            gen = torch.Generator(device=self.device)
            gen.manual_seed(format_pseudo_rn_seed(self.kwargs.get("hh_init_seed", None)))
            V = torch.clamp_min(
                torch.rand(
                    (batch_size, int(self.output_size)),
                    device=self.device,
                    dtype=torch.float32,
                    requires_grad=True,
                    generator=gen,
                ) * V_std + V_mu, min=0.0
            )
            I = torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            )
            Z = self.spike_func.apply(V, self.threshold, self.gamma)
            V = V * (1.0 - Z)
            return tuple([V, I, Z])
        elif self.kwargs["hh_init"] == "inputs":
            assert "inputs" in kwargs, "The inputs must be provided to initialize the state."
            assert int(self.input_size) == int(self.output_size), \
                "The input and output size must be the same with inputs initialization."
            # V_mu, V_std = self.kwargs.get("hh_init_mu", thr / 2.0), self.kwargs.get("hh_init_std", 0.341 * thr)
            V_mu, V_std = self.kwargs.get("hh_init_mu", 0.0), self.kwargs.get("hh_init_std", thr)
            gen = torch.Generator(device=self.device)
            gen.manual_seed(format_pseudo_rn_seed(self.kwargs.get("hh_init_seed", None)))
            I = torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            )
            Z = kwargs["inputs"].clone()
            V = (torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            ) * V_std + V_mu)
            V = (self.beta * V + self.alpha * I) * (1.0 - Z)

            return tuple([V, I, Z])
        return super(SpyLIFLayer, self).create_empty_state(batch_size=batch_size, **kwargs)

    def reset_regularization_loss(self):
        super(SpyLIFLayer, self).reset_regularization_loss()
        self._regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.device)
        self._regularization_l1 = torch.tensor(0.0, dtype=torch.float32, device=self.device)
        self._n_spike_per_neuron = torch.zeros(int(self.output_size), dtype=torch.float32, device=self.device)
        self._total_count = 0

    def update_regularization_loss(self, state: Optional[Any] = None, *args, **kwargs) -> torch.Tensor:
        """
        Update the regularization loss for this layer. Each update call increments the regularization loss so at the end
        the regularization loss will be the sum of all calls to this function.

        :param state: The current state of the layer.
        :return: The updated regularization loss.
        """
        next_V, next_I_syn, next_Z = state
        self._regularization_l1 += self.kwargs["spikes_regularization_factor"] * torch.sum(next_Z)
        # self._n_spike_per_neuron += torch.sum(torch.sum(next_Z, dim=0), dim=0)
        # self._total_count += next_Z.shape[0]*next_Z.shape[1]
        # current_l2 = self.kwargs["spikes_regularization_factor"]*torch.sum(self._n_spike_per_neuron ** 2) / (self._total_count + 1e-6)
        # self._regularization_loss = self._regularization_l1 + current_l2
        self._regularization_loss = self._regularization_l1
        return self._regularization_loss

    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        assert inputs.ndim == 2, f"Inputs must be of shape (batch_size, input_size), got {inputs.shape}."
        batch_size, nb_features = inputs.shape
        V, I_syn, Z = self._init_forward_state(state, batch_size, inputs=inputs)
        input_current = torch.matmul(inputs, self.forward_weights)
        if self.use_recurrent_connection:
            rec_current = torch.matmul(Z, torch.mul(self.recurrent_weights, self.rec_mask))
        else:
            rec_current = 0.0
        next_I_syn = self.alpha * I_syn + input_current + rec_current
        next_V = (self.beta * V + next_I_syn) * (1.0 - Z.detach())
        next_Z = self.spike_func.apply(next_V, self.threshold, self.gamma)
        return next_Z, (next_V, next_I_syn, next_Z)


class SpyALIFLayer(SpyLIFLayer):
    """
    The SpyALIF dynamic, inspired by Bellec and \\textit{al.} :cite:t:`bellec_solution_2020` and bye the
    :class:`SpyLIFLayer` from the work of Neftci :cite:t:`neftci_surrogate_2019`, is very
    similar to the SpyLIF dynamics (class :class:`SpyLIFLayer`). In fact, SpyALIF has exactly the same potential
    update equation as SpyLIF. The difference comes
    from the fact that the threshold potential varies with time and neuron input. Indeed, the threshold
    is increased at each output spike and is then decreased with a certain rate in order to come back to
    its starting threshold :math:`V_{\\text{th}}`. The threshold equation from :class:`SpyLIFLayer` is thus slightly
    modified by changing :math:`V_{\\text{th}} \\to A_j^t`. Thus, the output of neuron :math:`j` at time :math:`t`
    denoted :math:`z_j^t` is redefined by the equation :eq:`alif_z`.

    .. math::
        :label: SpyALIF_I

        \\begin{equation}
            I_{\\text{syn}, j}^{t+\\Delta t} = \\alpha I_{\text{syn}, j}^{t} + \\sum_{i}^{N} W_{ij}^{\\text{rec}} z_i^t
            + \\sum_i^{N} W_{ij}^{\\text{in}} x_i^{t+\\Delta t}
        \\end{equation}


    .. math::
        :label: SpyALIF_V

        \\begin{equation}
            V_j^{t+\\Delta t} = \\left(\\beta V_j^t + I_{\\text{syn}, j}^{t+\\Delta t}\\right) \\left(1 - z_j^t\\right)
        \\end{equation}


    .. math::
        :label: spyalif_alpha

        \\begin{equation}
            \\alpha = e^{-\\frac{\\Delta t}{\\tau_{\\text{syn}}}}
        \\end{equation}

    with :math:`\\tau_{\\text{syn}}` being the decay time constant of the synaptic current.

    .. math::
        :label: spyalif_beta

        \\begin{equation}
            \\beta = e^{-\\frac{\\Delta t}{\\tau_{\\text{mem}}}}
        \\end{equation}

    with :math:`\\tau_{\\text{syn}}` being the decay time constant of the synaptic current.

    The output of neuron :math:`j` at time :math:`t` denoted :math:`z_j^t` is defined by the equation :eq:`spyalif_z` .

    .. math::
        :label: spyalif_z

        z_j^t = H(V_j^t - A_j^t)

    where :math:`A_j^t` denotes the activation threshold of the neuron and the function :math:`H(\\cdot)`
    is the Heaviside function defined as :math:`H(x) = 1` if :math:`x \\geq 0` and :math:`H(x) = 0` otherwise.
    The update of the activation threshold is then described by :eq:`alif_A`.

    .. math::
        :label: alif_A

        \\begin{equation}
            A_j^t = V_{\\text{th}} + \\kappa a_j^t
        \\end{equation}

    with the adaptation variable :math:`a_j^t` described by :eq:`alif_a` and :math:`\\kappa` an amplification
    factor greater than 1 and typically equivalent to :math:`\\kappa\\approx 1.6` :cite:t:`bellec_solution_2020`.

    .. math::
        :label: alif_a

        \\begin{equation}
            a_j^{t+1} = \\rho a_j + z_j^t
        \\end{equation}

    With the decay factor :math:`\\rho` as:

    .. math::
        :label: alif_rho

        \\begin{equation}
            \\rho = e^{-\\frac{\\Delta t}{\\tau_a}}
        \\end{equation}

    SpyTorch library: https://github.com/surrogate-gradient-learning/spytorch.

    The variables of the equations :eq:`SpyALIF_I` and :eq:`SpyALIF_V` are described by the following definitions:

        - :math:`N` is the number of neurons in the layer.
        - :math:`I_{\\text{syn}, j}^{t}` is the synaptic current of neuron :math:`j` at time :math:`t`.
        - :math:`V_j^t` is the synaptic potential of the neuron :math:`j` at time :math:`t`.
        - :math:`\\Delta t` is the integration time step.
        - :math:`z_j^t` is the spike of the neuron :math:`j` at time :math:`t`.
        - :math:`\\alpha` is the decay constant of the synaptic current over time (equation :eq:`spylif_alpha`).
        - :math:`\\beta` is the decay constant of the membrane potential over time (equation :eq:`spylif_beta`).
        - :math:`W_{ij}^{\\text{rec}}` is the recurrent weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`W_{ij}^{\\text{in}}` is the input weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`x_i^{t}` is the input of the neuron :math:`i` at time :math:`t`.

    :Attributes:
        - :attr:`alpha` (torch.nn.Parameter): Decay constant of the synaptic current over time (equation :eq:`spyalif_alpha`).
        - :attr:`beta` (torch.nn.Parameter): Decay constant of the membrane potential over time (equation :eq:`spyalif_beta`).
        - :attr:`threshold` (torch.nn.Parameter): Activation threshold of the neuron (:math:`V_{\\text{th}}`).
        - :attr:`gamma` (torch.nn.Parameter): Slope of the Heaviside function (:math:`\\gamma`).
        - :attr:`kappa`: The amplification factor of the threshold potential (:math:`\\kappa`).
        - :attr:`rho`: The decay factor of the adaptation variable (:math:`\\rho`).
    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            use_recurrent_connection: bool = True,
            use_rec_eye_mask: bool = False,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        """
        Constructor for the SpyLIF layer.

        :param input_size: The size of the input.
        :type input_size: Optional[SizeTypes]
        :param output_size: The size of the output.
        :type output_size: Optional[SizeTypes]
        :param name: The name of the layer.
        :type name: Optional[str]
        :param use_recurrent_connection: Whether to use the recurrent connection.
        :type use_recurrent_connection: bool
        :param use_rec_eye_mask: Whether to use the recurrent eye mask.
        :type use_rec_eye_mask: bool
        :param spike_func: The spike function to use.
        :type spike_func: Callable[[torch.Tensor], torch.Tensor]
        :param learning_type: The learning type to use.
        :type learning_type: LearningType
        :param dt: Time step (Euler's discretisation).
        :type dt: float
        :param device: The device to use.
        :type device: Optional[torch.device]
        :param kwargs: The keyword arguments for the layer.

        :keyword float tau_syn: The synaptic time constant :math:`\\tau_{\\text{syn}}`. Default: 5.0 * dt.
        :keyword float tau_mem: The membrane time constant :math:`\\tau_{\\text{mem}}`. Default: 10.0 * dt.
        :keyword float threshold: The threshold potential :math:`V_{\\text{th}}`. Default: 1.0.
        :keyword float gamma: The multiplier of the derivative of the spike function :math:`\\gamma`. Default: 100.0.
        :keyword float spikes_regularization_factor: The regularization factor for the spikes. Higher this factor is,
            the more the network will tend to spike less. Default: 0.0.

        """
        self.spike_func = HeavisideSigmoidApprox
        super(SpyALIFLayer, self).__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=use_recurrent_connection,
            use_rec_eye_mask=use_rec_eye_mask,
            dt=dt,
            device=device,
            **kwargs
        )

        self.kappa = nn.Parameter(
            torch.tensor(self.kwargs["kappa"], dtype=torch.float32, device=self.device),
            requires_grad=self.kwargs["learn_kappa"]
        )
        self.rho = nn.Parameter(
            torch.tensor(np.exp(-dt / self.kwargs["tau_a"]), dtype=torch.float32, device=self.device),
            requires_grad=False
        )

    def _set_default_kwargs(self):
        self.kwargs.setdefault("tau_syn", 5.0 * self.dt)
        self.kwargs.setdefault("tau_mem", 10.0 * self.dt)
        self.kwargs.setdefault("tau_a", 200.0 * self.dt)
        self.kwargs.setdefault("threshold", 1.0)
        self.kwargs.setdefault("gamma", 100.0)
        self.kwargs.setdefault("kappa", 1.6)
        self.kwargs.setdefault("learn_kappa", False)
        self.kwargs.setdefault("spikes_regularization_factor", 0.0)
        self.kwargs.setdefault("hh_init", "zeros")

    def initialize_weights_(self):
        super().initialize_weights_()
        weight_scale = 0.2
        if self.kwargs.get("forward_weights", None) is not None:
            self.forward_weights.data = to_tensor(self.kwargs["forward_weights"]).to(self.device)
        else:
            torch.nn.init.normal_(self.forward_weights, mean=0.0, std=weight_scale / np.sqrt(int(self.input_size)))

        if self.kwargs.get("recurrent_weights", None) is not None and self.use_recurrent_connection:
            self.recurrent_weights.data = to_tensor(self.kwargs["recurrent_weights"]).to(self.device)
        elif self.use_recurrent_connection:
            torch.nn.init.normal_(self.recurrent_weights, mean=0.0, std=weight_scale / np.sqrt(int(self.output_size)))

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        """
        Create an empty state in the following form:
            ([membrane potential of shape (batch_size, self.output_size)],
            [synaptic current of shape (batch_size, self.output_size)],
            [spikes of shape (batch_size, self.output_size)])

        :param batch_size: The size of the current batch.
        :return: The current state.
        """
        kwargs.setdefault("n_hh", 4)
        thr = self.threshold.detach().cpu().item()
        if self.kwargs["hh_init"] == "random":
            V_mu, V_std = self.kwargs.get("hh_init_mu", thr / 2.0), self.kwargs.get("hh_init_std", 0.341 * thr)
            gen = torch.Generator(device=self.device)
            gen.manual_seed(format_pseudo_rn_seed(self.kwargs.get("hh_init_seed", None)))
            V = torch.clamp_min(
                torch.rand(
                    (batch_size, int(self.output_size)),
                    device=self.device,
                    dtype=torch.float32,
                    requires_grad=True,
                    generator=gen,
                ) * V_std + V_mu, min=0.0
            )
            I = torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            )
            Z = self.spike_func.apply(V, self.threshold, self.gamma)
            V = V * (1.0 - Z)
            a = torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            )
            return tuple([V, I, a, Z])
        elif self.kwargs["hh_init"] == "inputs":
            # V_mu, V_std = self.kwargs.get("hh_init_mu", thr / 2.0), self.kwargs.get("hh_init_std", 0.341 * thr)
            V_mu, V_std = self.kwargs.get("hh_init_mu", 0.0), self.kwargs.get("hh_init_std", thr)
            gen = torch.Generator(device=self.device)
            gen.manual_seed(format_pseudo_rn_seed(self.kwargs.get("hh_init_seed", None)))
            I = torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            )
            Z = kwargs["inputs"].clone()
            V = (torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            ) * V_std + V_mu)
            V = (self.beta * V + self.alpha * I) * (1.0 - Z)
            a = self.rho * torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            ) * thr + Z
            return tuple([V, I, a, Z])
        return super(SpyLIFLayer, self).create_empty_state(batch_size=batch_size, **kwargs)

    def reset_regularization_loss(self):
        super(SpyLIFLayer, self).reset_regularization_loss()
        self._regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.device)
        self._regularization_l1 = torch.tensor(0.0, dtype=torch.float32, device=self.device)
        self._n_spike_per_neuron = torch.zeros(int(self.output_size), dtype=torch.float32, device=self.device)
        self._total_count = 0

    def update_regularization_loss(self, state: Optional[Any] = None, *args, **kwargs) -> torch.Tensor:
        """
        Update the regularization loss for this layer. Each update call increments the regularization loss so at the end
        the regularization loss will be the sum of all calls to this function.

        :param state: The current state of the layer.
        :return: The updated regularization loss.
        """
        next_V, next_I_syn, next_Z = state
        self._regularization_l1 += self.kwargs["spikes_regularization_factor"] * torch.sum(next_Z)
        # self._n_spike_per_neuron += torch.sum(torch.sum(next_Z, dim=0), dim=0)
        # self._total_count += next_Z.shape[0]*next_Z.shape[1]
        # current_l2 = self.kwargs["spikes_regularization_factor"]*torch.sum(self._n_spike_per_neuron ** 2) / (self._total_count + 1e-6)
        # self._regularization_loss = self._regularization_l1 + current_l2
        self._regularization_loss = self._regularization_l1
        return self._regularization_loss

    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        assert inputs.ndim == 2, f"Inputs must be of shape (batch_size, input_size), got {inputs.shape}."
        batch_size, nb_features = inputs.shape
        V, I_syn, a, Z = self._init_forward_state(state, batch_size, inputs=inputs)
        input_current = torch.matmul(inputs, self.forward_weights)
        if self.use_recurrent_connection:
            rec_current = torch.matmul(Z, torch.mul(self.recurrent_weights, self.rec_mask))
        else:
            rec_current = 0.0
        next_I_syn = self.alpha * I_syn + input_current + rec_current
        next_V = (self.beta * V + next_I_syn) * (1.0 - Z.detach())
        next_a = self.rho * a + Z  # a^{t+1} = \rho * a_j^t + z_j^t
        A = self.threshold + self.kappa * next_a  # A_j^t = v_{th} + \kappa * a_j^t
        next_Z = self.spike_func.apply(next_V, A, self.gamma)  # z_j^t = H(v_j^t - A_j^t)
        return next_Z, (next_V, next_I_syn, next_a, next_Z)


# @inherit_fields_docstring(fields=["Attributes"], bases=[LIFLayer])
class ALIFLayer(LIFLayer):
    """
    The ALIF dynamic, inspired by Bellec and \\textit{al.} :cite:t:`bellec_solution_2020`, is very
    similar to the LIF dynamics (class :class:`LIFLayer`). In fact, ALIF has exactly the same potential
    update equation as LIF. The difference comes
    from the fact that the threshold potential varies with time and neuron input. Indeed, the threshold
    is increased at each output pulse and is then decreased with a certain rate in order to come back to
    its starting threshold :math:`V_{\\text{th}}`. The threshold equation from :class:`LIFLayer` is thus slightly
    modified by changing :math:`V_{\\text{th}} \\to A_j^t`. Thus, the output of neuron :math:`j` at time :math:`t`
    denoted :math:`z_j^t` is redefined by the equation :eq:`alif_z`.

    .. math::
        :label: alif_z

        \\begin{equation}
            z_j^t = H(V_j^t - A_j^t)
        \\end{equation}

    The update of the activation threshold is then described by :eq:`alif_A`.

    .. math::
        :label: alif_A

        \\begin{equation}
            A_j^t = V_{\\text{th}} + \\beta a_j^t
        \\end{equation}

    with the adaptation variable :math:`a_j^t` described by :eq:`alif_a` and :math:`\\beta` an amplification
    factor greater than 1 and typically equivalent to :math:`\\beta\\approx 1.6` :cite:t:`bellec_solution_2020`.

    .. math::
        :label: alif_a

        \\begin{equation}
            a_j^{t+1} = \\rho a_j + z_j^t
        \\end{equation}

    With the decay factor :math:`\\rho` as:

    .. math::
        :label: alif_rho

        \\begin{equation}
            \\rho = e^{-\\frac{\\Delta t}{\\tau_a}}
        \\end{equation}

    :Attributes:
        - :attr:`beta`: The amplification factor of the threshold potential :math:`\\beta`.
        - :attr:`rho`: The decay factor of the adaptation variable :math:`\\rho`.

    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            use_recurrent_connection: bool = True,
            use_rec_eye_mask: bool = False,
            spike_func: Type[SpikeFunction] = HeavisideSigmoidApprox,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        super(ALIFLayer, self).__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=use_recurrent_connection,
            use_rec_eye_mask=use_rec_eye_mask,
            spike_func=spike_func,
            dt=dt,
            device=device,
            **kwargs
        )
        self.beta = nn.Parameter(
            torch.tensor(self.kwargs["beta"], dtype=torch.float32, device=self.device),
            requires_grad=self.kwargs["learn_beta"]
        )
        self.rho = nn.Parameter(
            torch.tensor(np.exp(-dt / self.kwargs["tau_a"]), dtype=torch.float32, device=self.device),
            requires_grad=False
        )

    def _set_default_kwargs(self):
        self.kwargs.setdefault("tau_m", 20.0 * self.dt)
        self.kwargs.setdefault("tau_a", 200.0 * self.dt)
        self.kwargs.setdefault("beta", 1.6)
        # self.kwargs.setdefault("threshold", 0.03)
        self.kwargs.setdefault("threshold", 1.0)
        if issubclass(self.spike_func, HeavisideSigmoidApprox):
            self.kwargs.setdefault("gamma", 100.0)
        else:
            self.kwargs.setdefault("gamma", 0.3)
        self.kwargs.setdefault("learn_beta", False)
        self.kwargs.setdefault("spikes_regularization_factor", 0.0)

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        """
        Create an empty state in the following form:
            [[membrane potential of shape (batch_size, self.output_size)]
            [current threshold of shape (batch_size, self.output_size)]
            [spikes of shape (batch_size, self.output_size)]]

        :param batch_size: The size of the current batch.
        :type batch_size: int

        :return: The current state.
        :rtype: Tuple[torch.Tensor, ...]
        """
        kwargs.setdefault("n_hh", 3)
        return super(ALIFLayer, self).create_empty_state(batch_size=batch_size, **kwargs)

    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        assert inputs.ndim == 2
        batch_size, nb_features = inputs.shape
        V, a, Z = self._init_forward_state(state, batch_size, inputs=inputs)
        input_current = torch.matmul(inputs, self.forward_weights)
        if self.use_recurrent_connection:
            rec_current = torch.matmul(Z, torch.mul(self.recurrent_weights, self.rec_mask))
        else:
            rec_current = 0.0
        # v_j^{t+1} = \alpha * v_j^t + \sum_i W_{ji}*z_i^t + \sum_i W_{ji}^{in}x_i^{t+1} - z_j^t * v_{th}
        next_V = (self.alpha * V + input_current + rec_current) * (1.0 - Z.detach())
        next_a = self.rho * a + Z  # a^{t+1} = \rho * a_j^t + z_j^t
        A = self.threshold + self.beta * next_a  # A_j^t = v_{th} + \beta * a_j^t
        next_Z = self.spike_func.apply(next_V, A, self.gamma)  # z_j^t = H(v_j^t - A_j^t)
        return next_Z, (next_V, next_a, next_Z)

    def update_regularization_loss(self, state: Optional[Any] = None, *args, **kwargs) -> torch.Tensor:
        """
        Update the regularization loss for this layer. Each update call increments the regularization loss so at the end
        the regularization loss will be the sum of all calls to this function.

        :param state: The current state of the layer.
        :type state: Optional[Any]

        :return: The updated regularization loss.
        :rtype: torch.Tensor
        """
        next_V, next_a, next_Z = state
        self._regularization_loss += self.kwargs["spikes_regularization_factor"] * torch.sum(next_Z)
        # self._regularization_loss += 2e-6*torch.mean(torch.sum(next_Z, dim=-1)**2)
        return self._regularization_loss


class BellecLIFLayer(LIFLayer):
    """
    Layer implementing the LIF neuron model from the paper:
        "A solution to the learning dilemma for recurrent networks of spiking neurons"
        by Bellec et al. (2020) :cite:t:`bellec_solution_2020`.
    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            use_recurrent_connection: bool = True,
            use_rec_eye_mask: bool = True,
            spike_func: Type[SpikeFunction] = HeavisidePhiApprox,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        super().__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=use_recurrent_connection,
            use_rec_eye_mask=use_rec_eye_mask,
            spike_func=spike_func,
            dt=dt,
            device=device,
            **kwargs
        )

    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        assert inputs.ndim == 2
        batch_size, nb_features = inputs.shape
        V, Z = self._init_forward_state(state, batch_size, inputs=inputs)
        input_current = torch.matmul(inputs, self.forward_weights)
        if self.use_recurrent_connection:
            rec_current = torch.matmul(Z, torch.mul(self.recurrent_weights, self.rec_mask))
        else:
            rec_current = 0.0
        next_V = (self.alpha * V + input_current + rec_current) - Z.detach() * self.threshold
        next_Z = self.spike_func.apply(next_V, self.threshold, self.gamma)
        return next_Z, (next_V, next_Z)


# @inherit_fields_docstring(fields=["Attributes"], bases=[BaseNeuronsLayer])
class IzhikevichLayer(BaseNeuronsLayer):
    """
    Izhikevich p.274

    Not usable for now, stay tuned.
    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            use_recurrent_connection=True,
            use_rec_eye_mask=True,
            spike_func: Type[SpikeFunction] = HeavisideSigmoidApprox,
            dt=1e-3,
            device=None,
            **kwargs
    ):
        self.spike_func = spike_func
        super(IzhikevichLayer, self).__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=use_recurrent_connection,
            use_rec_eye_mask=use_rec_eye_mask,
            dt=dt,
            device=device,
            **kwargs
        )

        self.C = torch.tensor(self.kwargs["C"], dtype=torch.float32, device=self._device)
        self.v_rest = torch.tensor(self.kwargs["v_rest"], dtype=torch.float32, device=self._device)
        self.v_th = torch.tensor(self.kwargs["v_th"], dtype=torch.float32, device=self._device)
        self.k = torch.tensor(self.kwargs["k"], dtype=torch.float32, device=self._device)
        self.a = torch.tensor(self.kwargs["a"], dtype=torch.float32, device=self._device)
        self.b = torch.tensor(self.kwargs["b"], dtype=torch.float32, device=self._device)
        self.c = torch.tensor(self.kwargs["c"], dtype=torch.float32, device=self._device)
        self.d = torch.tensor(self.kwargs["d"], dtype=torch.float32, device=self._device)
        self.v_peak = torch.tensor(self.kwargs["v_peak"], dtype=torch.float32, device=self._device)
        self.gamma = torch.tensor(self.kwargs["gamma"], dtype=torch.float32, device=self._device)
        self.initialize_weights_()

    def _set_default_kwargs(self):
        self.kwargs.setdefault("C", 100.0)
        self.kwargs.setdefault("v_rest", -60.0)
        self.kwargs.setdefault("v_th", -40.0)
        self.kwargs.setdefault("k", 0.7)
        self.kwargs.setdefault("a", 0.03)
        self.kwargs.setdefault("b", -2.0)
        self.kwargs.setdefault("c", -50.0)
        self.kwargs.setdefault("d", 100.0)
        self.kwargs.setdefault("v_peak", 35.0)
        if isinstance(self.spike_func, HeavisideSigmoidApprox):
            self.kwargs.setdefault("gamma", 100.0)
        else:
            self.kwargs.setdefault("gamma", 1.0)

    def initialize_weights_(self):
        super().initialize_weights_()
        gain = 1.0
        for param in self.parameters():
            if param.ndim > 2:
                torch.nn.init.xavier_normal_(param, gain=gain)
            else:
                torch.nn.init.normal_(param, std=gain)

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        """
        Create an empty state in the following form:
            ([membrane potential of shape (batch_size, self.output_size)],
            [membrane potential of shape (batch_size, self.output_size)],
            [spikes of shape (batch_size, self.output_size)])
        :param batch_size: The size of the current batch.
        :return: The current state.
        """
        V = self.v_rest * torch.ones(
            (batch_size, int(self._output_size)),
            device=self._device,
            dtype=torch.float32,
            requires_grad=True,
        )
        u = torch.zeros(
            (batch_size, int(self._output_size)),
            device=self._device,
            dtype=torch.float32,
            requires_grad=True,
        )
        Z = torch.zeros(
            (batch_size, int(self._output_size)),
            device=self._device,
            dtype=torch.float32,
            requires_grad=True,
        )
        return V, u, Z

    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        assert inputs.ndim == 2
        batch_size, nb_features = inputs.shape
        V, u, Z = self._init_forward_state(state, batch_size)
        input_current = torch.matmul(inputs, self.forward_weights)
        if self.use_recurrent_connection:
            rec_current = torch.matmul(Z, torch.mul(self.recurrent_weights, self.rec_mask))
        else:
            rec_current = 0.0
        is_reset = Z.detach()
        I = input_current + rec_current
        dVdt = self.k * (V - self.v_rest) * (V - self.v_th) - u + I
        next_V = (V + self.dt * dVdt / self.C) * (1.0 - is_reset) + self.c * is_reset
        dudt = self.a * (self.b * (V - self.v_rest) - u)
        next_u = (u + self.dt * dudt) + self.d * is_reset
        next_Z = self.spike_func.apply(next_V, self.v_peak, self.gamma)
        return next_Z, (next_V, next_u, next_Z)



import enum
import os

import numpy as np
import torch
from torchvision.transforms import Compose, ToTensor


class ImgToSpikes:
    def __init__(
            self,
            n_steps: int,
            t_max: float = None,
            tau=20.0 * 1e-3,
            thr=0.2,
            use_periods=False,
            epsilon=1e-7,
    ):
        """
        :param n_steps: The number of time step
        :param t_max:
        :param tau: The membrane time constant of the LIF neuron to be charged
        :param thr: The firing threshold value
        :param epsilon: A generic (small) epsilon > 0
        """
        self.n_steps = n_steps
        self.t_max = n_steps if t_max is None else t_max
        self.tau = tau
        self.thr = thr
        self.epsilon = epsilon
        self.spikes_indices = None
        self.use_periods = use_periods
        self.spikes_gen_func = self.firing_periods_to_spikes if use_periods else self.firing_times_to_spikes

    def pixels_to_firing_periods(self, x: np.ndarray) -> np.ndarray:
        """
        Computes first firing time latency for a current input x assuming the charge time of a current based LIF neuron.
        :param x: the normalized between 0.0 and 1.0 input pixels
        :return: Time between spikes for each pixel of x
        """
        # t_per_step = self.t_max / self.n_steps
        idx = x < self.thr
        x = np.clip(x, self.thr + self.epsilon, 1.0e9)
        T = self.tau * np.log(x / (x - self.thr))
        T[idx] = self.t_max
        # periods = T // t_per_step
        return T.astype(int)

    def firing_periods_to_spikes_loop(self, firing_periods: np.ndarray) -> np.ndarray:
        spikes = np.zeros((self.n_steps, *firing_periods.shape), dtype=float)
        starts = np.clip(firing_periods, 0, self.n_steps - 1).astype(int)
        for i, period in enumerate(firing_periods):
            spikes[np.arange(starts[i], self.n_steps, step=period, dtype=int), i] = 1.0
        return spikes

    def firing_periods_to_spikes_clip(self, firing_periods: np.ndarray) -> np.ndarray:
        if self.spikes_indices is None:
            self.spikes_indices = np.indices((self.n_steps, *firing_periods.shape))
        starts = np.clip(firing_periods, 0, self.n_steps - 1, dtype=int)
        # starts[starts > (self.n_steps - 1)] = self.n_steps - 1
        spikes_range = self.spikes_indices[0] - starts[self.spikes_indices[1]]
        spikes = ((spikes_range % firing_periods[self.spikes_indices[1]]) == 0) * (spikes_range >= 0)
        return spikes.astype(float)

    def firing_periods_to_spikes(self, firing_periods: np.ndarray) -> np.ndarray:
        if self.spikes_indices is None:
            self.spikes_indices = np.indices((self.n_steps, *firing_periods.shape))
        firing_periods[firing_periods > (self.n_steps - 1)] = self.n_steps - 1
        firing_periods[firing_periods < 1] = 1
        spikes_range = self.spikes_indices[0] - firing_periods[self.spikes_indices[1]]
        spikes = ((spikes_range % firing_periods[self.spikes_indices[1]]) == 0) * (spikes_range >= 0)
        return spikes.astype(float)

    def firing_times_to_spikes(self, firing_times: np.ndarray) -> np.ndarray:
        spikes = np.zeros((self.n_steps, *firing_times.shape))
        firing_times_mask = firing_times < self.n_steps
        pix_indexes_masked = np.arange(len(firing_times))[firing_times_mask]
        spikes[firing_times[firing_times_mask], pix_indexes_masked] = 1.
        return spikes

    def _format_inputs(self, x) -> np.ndarray:
        if isinstance(x, torch.Tensor):
            return x.numpy()
        return x

    def __call__(self, x) -> torch.Tensor:
        x = self._format_inputs(x)
        firing_periods: np.ndarray = self.pixels_to_firing_periods(x)
        spikes = self.spikes_gen_func(firing_periods)
        return torch.tensor(spikes)




import warnings
from typing import Optional, Sequence, Union, Dict, Callable, List, Any

import numpy as np
import scipy
import torch

from .agent import Agent
from .buffers import BatchExperience, Experience
from .utils import discounted_cumulative_sums, continuous_actions_distribution
from ..transforms.base import to_numpy, to_tensor
from ..learning_algorithms.learning_algorithm import LearningAlgorithm
from ..utils import maybe_apply_softmax


class PPO(LearningAlgorithm):
    r"""
    Apply the Proximal Policy Optimization algorithm to the given model. The algorithm is described in the paper
    `Proximal Policy Optimization Algorithms <https://arxiv.org/abs/1707.06347>`.
    """
    CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"

    def __init__(
            self,
            agent: Optional[Agent] = None,
            optimizer: Optional[torch.optim.Optimizer] = None,
            **kwargs
    ):
        """
        Constructor of the PPO algorithm.

        :param agent: The agent to train.
        :type agent: Agent
        :param optimizer: The optimizer to use.
        :type optimizer: torch.optim.Optimizer
        :param kwargs: Additional keyword arguments.

        :keyword float clip_ratio: The clipping ratio for the policy loss.
        :keyword float tau: The smoothing factor for the policy update.
        :keyword float gamma: The discount factor.
        :keyword float gae_lambda: The lambda parameter for the generalized advantage estimation (GAE).
        :keyword float critic_weight: The weight of the critic loss.
        :keyword float entropy_weight: The weight of the entropy loss.
        :keyword torch.nn.Module critic_criterion: The loss function to use for the critic.
        :keyword bool advantages=returns-values: This keyword is introduced to fix a bug when using the GAE. If set to
            True, the advantages are computed as the returns minus the values. If set to False, the advantages are
            compute as in the PPO paper. The default value is False and it is recommended to try to set it to True
            if the agent doesn't seem to learn.
        :keyword float max_grad_norm: The maximum L2 norm of the gradient. Default is 0.5.
        """
        kwargs.setdefault("save_state", True)
        kwargs.setdefault("load_state", True)
        super().__init__(params=None, **kwargs)
        self._agent = agent
        self.last_agent = None
        self.optimizer = optimizer
        self.critic_optimizer = None
        self.policy_params = None
        self.critic_params = None
        self.continuous_criterion = torch.nn.MSELoss()
        self.discrete_criterion = torch.nn.CrossEntropyLoss()
        self.clip_ratio = kwargs.get("clip_ratio", 0.2)
        self.critic_clip = kwargs.get("critic_clip", 0.2)
        self.tau = kwargs.get("tau", 0.0)
        self.gamma = kwargs.get("gamma", 0.99)
        self.gae_lambda = kwargs.get("gae_lambda", 0.99)
        self.critic_weight = kwargs.get("critic_weight", 0.5)
        self.entropy_weight = kwargs.get("entropy_weight", 0.01)
        self.critic_criterion = kwargs.get("critic_criterion", torch.nn.MSELoss())
        self.adv_as_returns_values = kwargs.get("advantages=returns-values", False)
        self.max_grad_norm = kwargs.get("max_grad_norm", 0.5)

    @property
    def policy(self):
        if self.agent is None:
            return None
        return self.agent.policy

    @property
    def critic(self):
        if self.agent is None:
            return None
        return self.agent.critic

    @property
    def last_policy(self):
        if self.last_agent is None:
            return None
        return self.last_agent.policy

    @last_policy.setter
    def last_policy(self, policy):
        self.last_agent.policy = policy

    @property
    def agent(self):
        if self._agent is not None:
            return self._agent
        if self.trainer is None:
            return None
        return self.trainer.agent

    def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
        if self.save_state:
            state = checkpoint.get(self.name, {})
            opt_state_dict = state.get(self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY, None)
            if opt_state_dict is not None:
                self.optimizer.load_state_dict(opt_state_dict)

    def get_checkpoint_state(self, trainer, **kwargs) -> object:
        if self.save_state:
            if self.optimizer is not None:
                return {
                    self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: self.optimizer.state_dict()
                }
        return None

    def start(self, trainer, **kwargs):
        super().start(trainer, **kwargs)
        self.policy_params = list(filter(lambda p: p.requires_grad, self.policy.parameters()))
        self.critic_params = list(filter(lambda p: p.requires_grad, self.critic.parameters()))
        self.params = list(self.policy_params + self.critic_params)
        param_groups = [
            {"params": self.policy_params, "lr": self.kwargs.get("default_policy_lr", 2e-4)},
            {"params": self.critic_params, "lr": self.kwargs.get("default_critic_lr", 1e-3)},
        ]
        if self.optimizer is None:
            self.optimizer = torch.optim.Adam(param_groups)

        self.last_agent = trainer.copy_agent()
        if self.tau is None:
            self.tau = 1 / trainer.state.n_epochs
        assert self.tau >= 0, "The parameter `tau` must be greater or equal to 0."

    def _compute_policy_distributions(self, batch: BatchExperience, **kwargs):
        """
        Computes the policy distributions for the given batch.

        :param batch: The batch to compute the policy distributions for.
        :param kwargs: Other keyword arguments.

        :return: The policy distributions and the last policy distributions.
        """
        obs_as_tensor = to_tensor(batch.obs)
        policy_preds = kwargs.get(
            "policy_preds", self.agent.get_actions(obs_as_tensor, re_format="raw", as_numpy=False)
        )
        with torch.no_grad():
            last_policy_preds = self.last_agent.get_actions(obs_as_tensor, re_format="raw", as_numpy=False)

        if isinstance(policy_preds, dict):
            policy_dist, last_policy_dist = {}, {}
            for k in policy_preds:
                if k in self.agent.discrete_actions:
                    policy_dist[k] = torch.distributions.Categorical(
                        probs=maybe_apply_softmax(policy_preds[k], dim=-1)
                    )
                    last_policy_dist[k] = torch.distributions.Categorical(
                        probs=maybe_apply_softmax(last_policy_preds[k], dim=-1)
                    )
                else:
                    # TODO: must get the right covariance for each continuous action, see :class:`Agent`.
                    covariance = self.agent.get_continuous_action_covariances()[self.agent.continuous_actions[0]]
                    policy_dist[k] = continuous_actions_distribution(policy_preds[k], covariance=covariance)
                    last_policy_dist[k] = continuous_actions_distribution(last_policy_preds[k], covariance=covariance)
        elif self.agent.discrete_actions:
            policy_dist = torch.distributions.Categorical(probs=maybe_apply_softmax(policy_preds, dim=-1))
            last_policy_preds_smax = maybe_apply_softmax(last_policy_preds, dim=-1)
            last_policy_dist = torch.distributions.Categorical(probs=last_policy_preds_smax)
        else:
            covariance = self.agent.get_continuous_action_covariances()[self.agent.continuous_actions[0]]
            policy_dist = continuous_actions_distribution(policy_preds, covariance=covariance)
            last_policy_dist = continuous_actions_distribution(last_policy_preds, covariance=covariance)
        return policy_dist, last_policy_dist

    def _compute_policy_ratio(self, batch: BatchExperience, **kwargs) -> torch.Tensor:
        obs_as_tensor = to_tensor(batch.obs)
        actions = self.get_actions_from_batch(batch)
        policy_preds = kwargs.get("policy_preds", self.agent.get_actions(obs_as_tensor, re_format="raw", as_numpy=False))
        with torch.no_grad():
            last_policy_preds = self.last_agent.get_actions(obs_as_tensor, re_format="raw", as_numpy=False)

        if isinstance(policy_preds, dict):
            policy_ratio = {}
            for k in policy_preds:
                key_actions = actions[k] if isinstance(actions, dict) else actions
                if k in self.agent.discrete_actions:
                    policy_dist = torch.distributions.Categorical(
                        probs=maybe_apply_softmax(policy_preds[k], dim=-1)
                    )
                    last_policy_dist = torch.distributions.Categorical(
                        probs=maybe_apply_softmax(last_policy_preds[k], dim=-1)
                    )
                else:
                    policy_dist = continuous_actions_distribution(policy_preds[k])
                    last_policy_dist = continuous_actions_distribution(last_policy_preds[k])
                # policy_ratio[k] = policy_preds[k] / (last_policy_preds[k] + 1e-8)
                policy_ratio[k] = torch.exp(
                    policy_dist.log_prob(key_actions) - last_policy_dist.log_prob(key_actions)
                )
        elif self.agent.discrete_actions:
            key_actions = actions[list(actions.keys())[0]] if isinstance(actions, dict) else actions
            policy_dist = torch.distributions.Categorical(probs=maybe_apply_softmax(policy_preds, dim=-1))
            last_policy_preds_smax = maybe_apply_softmax(last_policy_preds, dim=-1)
            last_policy_dist = torch.distributions.Categorical(probs=last_policy_preds_smax)
            policy_ratio = torch.exp(
                policy_dist.log_prob(key_actions) - last_policy_dist.log_prob(key_actions)
            )
        else:
            policy_dist = continuous_actions_distribution(policy_preds)
            last_policy_dist = continuous_actions_distribution(last_policy_preds)
            key_actions = actions[list(actions.keys())[0]] if isinstance(actions, dict) else actions
            policy_ratio = torch.exp(
                policy_dist.log_prob(key_actions) - last_policy_dist.log_prob(key_actions)
            )
        # policy_ratio = policy_preds / (last_policy_preds + 1e-8)
        return policy_ratio

    def _compute_policy_ratio_from_distributions(
            self, batch: BatchExperience, policy_dist, last_policy_dist, **kwargs
    ):
        actions = self.get_actions_from_batch(batch)
        if isinstance(policy_dist, dict):
            policy_ratio = {}
            for k in policy_dist:
                key_actions = actions[k] if isinstance(actions, dict) else actions
                policy_ratio[k] = torch.exp(
                    policy_dist[k].log_prob(key_actions) - last_policy_dist[k].log_prob(key_actions)
                )
        elif self.agent.discrete_actions:
            key_actions = actions[list(actions.keys())[0]] if isinstance(actions, dict) else actions
            policy_ratio = torch.exp(
                policy_dist.log_prob(key_actions) - last_policy_dist.log_prob(key_actions)
            )
        else:
            key_actions = actions[list(actions.keys())[0]] if isinstance(actions, dict) else actions
            policy_ratio = torch.exp(
                policy_dist.log_prob(key_actions) - last_policy_dist.log_prob(key_actions)
            )
        return policy_ratio

    def _compute_policy_loss(self, batch: BatchExperience, **kwargs) -> torch.Tensor:
        # policy_ratio = self._compute_policy_ratio(batch, **kwargs)
        policy_ratio = self._compute_policy_ratio_from_distributions(batch, **kwargs)
        advantages = self.get_advantages_from_batch(batch)
        if not isinstance(policy_ratio, dict):
            policy_ratio = {"default": policy_ratio}
        policy_loss = to_tensor(0.0).to(self.policy.device)
        for key, ratio in policy_ratio.items():
            view_shape = [policy_ratio[key].shape[0], ] + (policy_ratio[key].ndim - 1) * [1]
            ratio_adv = policy_ratio[key] * advantages.view(*view_shape).to(self.policy.device)
            ratio_clamped = torch.clamp(policy_ratio[key], min=1.0 - self.clip_ratio, max=1.0 + self.clip_ratio)
            ratio_adv_clamped = ratio_clamped * advantages.view(*view_shape).to(self.policy.device)
            policy_loss += -torch.mean(torch.min(ratio_adv, ratio_adv_clamped))
        return policy_loss

    def _compute_critic_loss(self, batch: BatchExperience, **kwargs) -> torch.Tensor:
        """
        Compute the critic loss.

        The math are taken from https://github.com/Unity-Technologies/ml-agents/blob/6bb711f1b0a29a06b4deb544bb1e93c392acb255/ml-agents/mlagents/trainers/torch_entities/utils.py#L400.

        :param batch:
        :param kwargs:
        :return:
        """
        critic_predictions = self.critic(to_tensor(batch.obs))
        with torch.no_grad():
            last_critic_preds = self.last_agent.critic(to_tensor(batch.obs))
        if isinstance(critic_predictions, dict):
            assert len(critic_predictions) == 1, "Only one critic output is supported."
            critic_values = critic_predictions[list(critic_predictions.keys())[0]].view(-1)
            last_critic_preds = last_critic_preds[list(last_critic_preds.keys())[0]].view(-1)
        else:
            critic_values = critic_predictions.view(-1)
            last_critic_preds = last_critic_preds.view(-1)
        values_targets = self.get_returns_from_batch(batch).view(-1)
        clipped_value_estimate = last_critic_preds + torch.clamp(
            critic_values - last_critic_preds, -1 * self.critic_clip, self.critic_clip
        )
        v_opt_a = self.critic_criterion(critic_values, values_targets)
        v_opt_b = self.critic_criterion(critic_values, clipped_value_estimate)
        critic_loss = torch.mean(torch.max(v_opt_a, v_opt_b))
        return critic_loss

    def _compute_entropy_loss(self, batch: BatchExperience, **kwargs) -> torch.Tensor:
        obs_as_tensor = to_tensor(batch.obs)
        policy_preds = kwargs.get("policy_preds", self.agent.get_actions(obs_as_tensor, re_format="raw", as_numpy=False))
        if isinstance(policy_preds, dict):
            entropy_loss = to_tensor(0.0).to(self.policy.device)
            for k in policy_preds:
                if k in self.agent.discrete_actions:
                    policy_dist = torch.distributions.Categorical(
                        probs=maybe_apply_softmax(policy_preds[k], dim=-1)
                    )
                    entropy_loss += torch.mean(policy_dist.entropy())
                else:
                    # TODO: implement covariance matrix to compute entropy
                    # policy_dist = torch.distributions.MultivariateNormal(policy_preds[k], torch.eye(policy_preds[k].shape[-1]))
                    # entropy_loss += torch.mean(policy_dist.entropy())
                    pass
        elif self.agent.discrete_actions:
            policy_dist = torch.distributions.Categorical(probs=maybe_apply_softmax(policy_preds, dim=-1))
            entropy_loss = torch.mean(policy_dist.entropy())
        else:
            # TODO: implement covariance matrix to compute entropy
            # policy_dist = torch.distributions.MultivariateNormal(policy_preds, torch.eye(policy_preds.shape[-1]))
            # entropy_loss = torch.mean(policy_dist.entropy())
            entropy_loss = to_tensor(0.0).to(self.policy.device)
        return entropy_loss

    def _compute_entropy_loss_from_distributions(
            self, batch: BatchExperience, **kwargs
    ) -> torch.Tensor:
        obs_as_tensor = to_tensor(batch.obs)
        policy_preds = kwargs.get(
            "policy_preds", self.agent.get_actions(obs_as_tensor, re_format="raw", as_numpy=False)
        )
        kwargs["policy_preds"] = policy_preds
        policy_dist = kwargs.get("policy_dist", self._compute_policy_distributions(batch, **kwargs)[0])
        if isinstance(policy_dist, dict):
            entropy_loss = to_tensor(0.0).to(self.policy.device)
            for k in policy_dist:
                entropy_loss += torch.mean(policy_dist[k].entropy())
        else:
            entropy_loss = torch.mean(policy_dist.entropy())
        return entropy_loss

    def update_params(self, batch: BatchExperience) -> float:
        """
        Performs a single update of the policy network using the provided optimizer and buffer
        """
        policy_preds = self.agent.get_actions(to_tensor(batch.obs), re_format="raw", as_numpy=False)
        policy_dist, last_policy_dist = self._compute_policy_distributions(batch, policy_preds=policy_preds)
        policy_loss = self._compute_policy_loss(
            batch, policy_preds=policy_preds, policy_dist=policy_dist, last_policy_dist=last_policy_dist
        )
        critic_loss = self._compute_critic_loss(batch)
        entropy_loss = self._compute_entropy_loss_from_distributions(
            batch, policy_preds=policy_preds, policy_dist=policy_dist, last_policy_dist=last_policy_dist
        )
        weighted_critic_loss = self.critic_weight * critic_loss.to(self.policy.device)
        weighted_entropy_loss = self.entropy_weight * entropy_loss.to(self.policy.device)
        loss = policy_loss + weighted_critic_loss - weighted_entropy_loss

        self.optimizer.zero_grad()
        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.params, self.max_grad_norm)
        self.optimizer.step()
        self.agent.decay_continuous_action_variances()

        return to_numpy(loss).item()

    def _batch_obs(self, batch: List[Experience]):
        as_dict = isinstance(batch[0].obs, dict)
        if as_dict:
            obs_batched = batch[0].obs
            for key in obs_batched:
                obs_batched[key] = torch.stack([to_tensor(ex.obs[key]) for ex in batch]).to(self.policy.device)
        else:
            obs_batched = torch.stack([to_tensor(ex.obs) for ex in batch]).to(self.policy.device)
        return obs_batched

    def _compute_advantages(self, trajectory, values):
        values = to_numpy(values).reshape(-1)
        if self.adv_as_returns_values:
            advantages = self._compute_returns(trajectory, values) - values
        else:
            terminals = np.asarray([ex.terminal for ex in trajectory])
            rewards = np.array([ex.reward for ex in trajectory.experiences])
            rewards = np.append(rewards, (1 - int(terminals[-1])) * values[-1])  # from: https://keras.io/examples/rl/ppo_cartpole/
            values = np.append(values, (1 - int(terminals[-1])) * values[-1])  # from: https://keras.io/examples/rl/ppo_cartpole/
            deltas = rewards[:-1] + self.gamma * values[1:] - values[:-1]
            advantages = discounted_cumulative_sums(deltas, self.gamma * self.gae_lambda)

        adv_mean, adv_std = np.mean(advantages), np.std(advantages)
        advantages = (advantages - adv_mean) / (adv_std + 1e-8)
        return advantages

    def _compute_values(self, trajectory):
        obs_as_tensor = BatchExperience(trajectory).obs
        values = self.agent.get_values(obs_as_tensor, as_numpy=True, re_as_dict=False).reshape(-1)
        return values

    def _compute_returns(self, trajectory, values):
        terminals = np.asarray([ex.terminal for ex in trajectory])
        values = to_numpy(values).reshape(-1)
        values = np.append(values, (1 - int(terminals[-1])) * values[-1])  # from: https://keras.io/examples/rl/ppo_cartpole/
        rewards = np.array([ex.reward for ex in trajectory.experiences])
        rewards = np.append(rewards, (1 - int(terminals[-1])) * values[-1])  # from: https://keras.io/examples/rl/ppo_cartpole/
        returns = discounted_cumulative_sums(rewards, self.gamma)[:-1]
        returns_mean, returns_std = returns.mean(), returns.std()
        returns = (returns - returns_mean) / (returns_std + 1e-8)
        return returns

    def get_advantages_from_batch(self, batch: BatchExperience) -> torch.Tensor:
        """
        Computes the advantages for the provided batch
        """
        assert all("advantage" in x for x in batch.others), "All experiences in the batch must have an advantage."
        advantages = to_tensor([x["advantage"] for x in batch.others]).to(self.policy.device)
        return advantages

    def get_values_from_batch(self, batch: BatchExperience) -> torch.Tensor:
        """
        Computes the values for the provided batch
        """
        assert all("value" in x for x in batch.others), "All experiences in the batch must have a value."
        values = to_tensor([x["value"] for x in batch.others]).to(self.policy.device)
        return values

    def get_returns_from_batch(self, batch: BatchExperience) -> torch.Tensor:
        """
        Computes the returns for the provided batch
        """
        assert all("return" in x for x in batch.others), "All experiences in the batch must have a return."
        returns = to_tensor([x["return"] for x in batch.others]).to(self.policy.device)
        return returns

    def get_actions_from_batch(self, batch: BatchExperience) -> torch.Tensor:
        """
        Get the actions for the provided batch
        """
        actions = batch.actions
        if isinstance(actions, dict):
            for key in actions:
                if key in self.agent.discrete_actions:
                    actions[key] = actions[key].to(self.policy.device)
                    if actions[key].dim() > 1 and actions[key].shape[-1] > 1:
                        actions[key] = torch.argmax(actions[key], dim=-1)
                    actions[key] = actions[key].long()
        elif self.agent.discrete_actions:
            if actions.dim() > 1 and actions.shape[-1] > 1:
                actions = torch.argmax(actions, dim=-1)
            actions = actions.long()
        return actions

    def on_optimization_begin(self, trainer, **kwargs):
        super().on_optimization_begin(trainer, **kwargs)
        batch = trainer.current_training_state.x_batch
        batch_loss = self.update_params(batch)
        trainer.update_state_(batch_loss=batch_loss)

    def on_optimization_end(self, trainer, **kwargs):
        super().on_optimization_end(trainer, **kwargs)
        if not np.isclose(self.tau, 0.0):
            self.last_policy.soft_update(self.policy, tau=self.tau)

    def on_iteration_begin(self, trainer, **kwargs):
        super().on_iteration_begin(trainer, **kwargs)
        self.last_policy.hard_update(self.policy)

    def on_trajectory_end(self, trainer, trajectory, **kwargs) -> List[Dict[str, Any]]:
        super().on_trajectory_end(trainer, trajectory, **kwargs)
        if len(trajectory.experiences) == 0:
            return []
        values = self._compute_values(trajectory)
        advantages = self._compute_advantages(trajectory, values)
        returns = self._compute_returns(trajectory, values)
        trajectory_metrics = [
            {"advantage": advantage, "value": value, "return": returns_item}
            for advantage, value, returns_item in zip(advantages, values, returns)
        ]
        trajectory.update_others(trajectory_metrics)
        # for i, exp in enumerate(trajectory.experiences):
        # 	exp.others.update(trajectory_metrics[i])
        # batch_loss = self.update_params(BatchExperience(trajectory.experiences, self.policy.device))
        # trainer.update_state_(batch_loss=batch_loss)
        return trajectory_metrics

    def on_pbar_update(self, trainer, **kwargs) -> dict:
        """
        Called when the progress bar is updated.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param kwargs: Additional arguments.

        :return: None
        """
        repr_action_var = {
            k: [float(f"{v:.3f}") for v in self.agent.continuous_action_variances[k].data.tolist()]
            for k in self.agent.continuous_action_variances
        }
        return {"actions_var": repr_action_var}
	



from typing import Optional, Tuple

import numpy as np
import torch
from torch import nn

from .base import BaseNeuronsLayer
from ...dimension import SizeTypes
from ...transforms import to_tensor


# @inherit_fields_docstring(fields=["Attributes"], bases=[BaseNeuronsLayer])
class LILayer(BaseNeuronsLayer):
    """
    The integration in time of these dynamics is done using the equation
    :eq:`li_v` inspired by Bellec and al. :cite:t:`bellec_solution_2020`.

    .. math::
        :label: li_v

        \\begin{equation}
            V_j^{t+\\Delta t} = \\kappa V_j^{t} + \\sum_{i}^N W_{ij}x_i^{t+\\Delta t} + b_j
        \\end{equation}

    .. math::
        :label: li_kappa

        \\begin{equation}
            \\kappa = e^{-\\frac{\\Delta t}{\\tau_{\\text{mem}}}}
        \\end{equation}


    The parameters of the equation :eq:`li_v` are:

        - :math:`N` is the number of neurons in the layer.
        - :math:`V_j^t` is the synaptic potential of the neuron :math:`j` at time :math:`t`.
        - :math:`\\Delta t` is the integration time step.
        - :math:`\\kappa` is the decay constant of the synaptic current over time (equation :eq:`li_kappa`).
        - :math:`W_{ij}^{\\text{rec}}` is the recurrent weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`W_{ij}^{\\text{in}}` is the input weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`x_i^{t}` is the input of the neuron :math:`i` at time :math:`t`.

    :Attributes:
        - :attr:`bias_weights` (torch.nn.Parameter): Bias weights of the layer.
        - :attr:`kappa` (torch.nn.Parameter): Decay constant of the synaptic current over time see equation :eq:`li_kappa`.

    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        super(LILayer, self).__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=False,
            dt=dt,
            device=device,
            **kwargs
        )
        self.bias_weights = None
        self.kappa = torch.nn.Parameter(
            torch.tensor(self.kwargs["kappa"], dtype=torch.float32, device=self.device),
            requires_grad=self.kwargs["learn_kappa"]
        )

    def _set_default_kwargs(self):
        self.kwargs.setdefault("tau_out", 10.0 * self.dt)
        self.kwargs.setdefault("kappa", np.exp(-self.dt / self.kwargs["tau_out"]))
        self.kwargs.setdefault("learn_kappa", False)
        self.kwargs.setdefault("use_bias", True)

    def build(self) -> 'LILayer':
        if self.kwargs["use_bias"]:
            self.bias_weights = nn.Parameter(
                torch.empty((int(self.output_size),), device=self._device),
                requires_grad=self.requires_grad,
            )
        else:
            self.bias_weights = torch.zeros((int(self.output_size),), dtype=torch.float32, device=self._device)
        super(LILayer, self).build()
        self.initialize_weights_()
        return self

    def initialize_weights_(self):
        super(LILayer, self).initialize_weights_()
        if self.kwargs.get("bias_weights", None) is not None:
            self.bias_weights.data = to_tensor(self.kwargs["bias_weights"]).to(self.device)
        else:
            torch.nn.init.constant_(self.bias_weights, 0.0)

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        """
        Create an empty state in the following form:
            [membrane potential of shape (batch_size, self.output_size)]

        :param batch_size: The size of the current batch.
        :return: The current state.
        """
        kwargs.setdefault("n_hh", 1)
        return super(LILayer, self).create_empty_state(batch_size=batch_size, **kwargs)

    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        assert inputs.ndim == 2
        batch_size, nb_features = inputs.shape
        V, = self._init_forward_state(state, batch_size, inputs=inputs)
        next_V = self.kappa * V + torch.matmul(inputs, self.forward_weights) + self.bias_weights
        return self.activation(next_V), (next_V,)

    def extra_repr(self) -> str:
        _repr = super(LILayer, self).extra_repr()
        _repr += f", bias={self.kwargs['use_bias']}"
        if self.kwargs['learn_kappa']:
            _repr += f", learn_kappa={self.kwargs['learn_kappa']}"
        else:
            _repr += f", kappa={self.kappa.item():.2f}"
        _repr += f", activation={self.activation.__class__.__name__}"
        return _repr


# @inherit_fields_docstring(fields=["Attributes"], bases=[BaseNeuronsLayer])
class SpyLILayer(BaseNeuronsLayer):
    """
    The SpyLI dynamics is a more complex variant of the LI dynamics (class :class:`LILayer`) allowing it to have a
    greater power of expression. This variant is also inspired by Neftci :cite:t:`neftci_surrogate_2019` and also
    contains  two differential equations like the SpyLIF dynamics :class:`SpyLIFLayer`. The equation :eq:`SpyLI_I`
    presents the synaptic current update equation with euler integration while the equation :eq:`SpyLI_V` presents the
    synaptic potential update.

    .. math::
        :label: SpyLI_I

        \\begin{equation}
            I_{\\text{syn}, j}^{t+\\Delta t} = \\alpha I_{\text{syn}, j}^{t} +
            \\sum_{i}^{N} W_{ij}^{\\text{rec}} I_{\\text{syn}, j}^{t}
            + \\sum_i^{N} W_{ij}^{\\text{in}} x_i^{t+\\Delta t}
        \\end{equation}


    .. math::
        :label: SpyLI_V

        \\begin{equation}
            V_j^{t+\\Delta t} = \\beta V_j^t + I_{\\text{syn}, j}^{t+\\Delta t} + b_j
        \\end{equation}


    .. math::
        :label: spyli_alpha

        \\begin{equation}
            \\alpha = e^{-\\frac{\\Delta t}{\\tau_{\\text{syn}}}}
        \\end{equation}

    with :math:`\\tau_{\\text{syn}}` being the decay time constant of the synaptic current.

    .. math::
        :label: spyli_beta

        \\begin{equation}
            \\beta = e^{-\\frac{\\Delta t}{\\tau_{\\text{mem}}}}
        \\end{equation}

    with :math:`\\tau_{\\text{syn}}` being the decay time constant of the synaptic current.

    SpyTorch library: https://github.com/surrogate-gradient-learning/spytorch.

    The variables of the equations :eq:`SpyLI_I` and :eq:`SpyLI_V` are described by the following definitions:

        - :math:`N` is the number of neurons in the layer.
        - :math:`I_{\\text{syn}, j}^{t}` is the synaptic current of neuron :math:`j` at time :math:`t`.
        - :math:`V_j^t` is the synaptic potential of the neuron :math:`j` at time :math:`t`.
        - :math:`\\Delta t` is the integration time step.
        - :math:`\\alpha` is the decay constant of the synaptic current over time (equation :eq:`spyli_alpha`).
        - :math:`\\beta` is the decay constant of the membrane potential over time (equation :eq:`spyli_beta`).
        - :math:`W_{ij}^{\\text{rec}}` is the recurrent weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`W_{ij}^{\\text{in}}` is the input weight of the neuron :math:`i` to the neuron :math:`j`.
        - :math:`x_i^{t}` is the input of the neuron :math:`i` at time :math:`t`.

    :Attributes:
        - :attr:`alpha` (torch.nn.Parameter): Decay constant of the synaptic current over time (equation :eq:`spyli_alpha`).
        - :attr:`beta` (torch.nn.Parameter): Decay constant of the membrane potential over time (equation :eq:`spyli_beta`).
        - :attr:`gamma` (torch.nn.Parameter): Slope of the Heaviside function (:math:`\\gamma`).
    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        super(SpyLILayer, self).__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            use_recurrent_connection=False,
            dt=dt,
            device=device,
            **kwargs
        )
        self.bias_weights = None
        self.alpha = torch.nn.Parameter(
            torch.tensor(self.kwargs["alpha"], dtype=torch.float32, device=self.device),
            requires_grad=self.kwargs["learn_alpha"]
        )
        self.beta = torch.nn.Parameter(
            torch.tensor(self.kwargs["beta"], dtype=torch.float32, device=self.device),
            requires_grad=self.kwargs["learn_beta"]
        )

    def _set_default_kwargs(self):
        self.kwargs.setdefault("tau_syn", 5.0 * self.dt)
        self.kwargs.setdefault("alpha", np.exp(-self.dt / self.kwargs["tau_syn"]))
        self.kwargs.setdefault("learn_alpha", False)
        self.kwargs.setdefault("tau_mem", 10.0 * self.dt)
        self.kwargs.setdefault("beta", np.exp(-self.dt / self.kwargs["tau_mem"]))
        self.kwargs.setdefault("learn_beta", False)
        self.kwargs.setdefault("use_bias", False)

    def build(self) -> 'SpyLILayer':
        super(SpyLILayer, self).build()
        if self.kwargs["use_bias"]:
            self.bias_weights = torch.nn.Parameter(
                torch.empty((int(self.output_size),), device=self._device),
                requires_grad=self.requires_grad,
            )
        self.initialize_weights_()
        return self

    def initialize_weights_(self):
        super(SpyLILayer, self).initialize_weights_()
        weight_scale = 0.2
        if self.kwargs.get("forward_weights", None) is not None:
            self.forward_weights.data = to_tensor(self.kwargs["forward_weights"]).to(self.device)
        else:
            torch.nn.init.normal_(self.forward_weights, mean=0.0, std=weight_scale / np.sqrt(int(self.input_size)))
        if self.kwargs["use_bias"]:
            if self.kwargs.get("bias_weights", None) is not None:
                self.bias_weights.data = to_tensor(self.kwargs["bias_weights"]).to(self.device)
            else:
                torch.nn.init.constant_(self.bias_weights, 0.0)

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        """
        Create an empty state in the following form:
            [membrane potential of shape (batch_size, self.output_size),
            synaptic current of shape (batch_size, self.output_size)]

        :param batch_size: The size of the current batch.
        :return: The current state.
        """
        kwargs.setdefault("n_hh", 2)
        return super(SpyLILayer, self).create_empty_state(batch_size=batch_size, **kwargs)

    def forward(
            self,
            inputs: torch.Tensor,
            state: Tuple[torch.Tensor, ...] = None,
            **kwargs
    ):
        assert inputs.ndim == 2
        batch_size, nb_features = inputs.shape
        V, I_syn = self._init_forward_state(state, batch_size, inputs=inputs)
        next_I_syn = self.alpha * I_syn + torch.matmul(inputs, self.forward_weights)
        if self.bias_weights is not None:
            next_V = self.beta * V + next_I_syn + self.bias_weights
        else:
            next_V = self.beta * V + next_I_syn
        return self.activation(next_V), (next_V, next_I_syn)

    def extra_repr(self) -> str:
        _repr = super().extra_repr()
        _repr += f", bias={self.kwargs['use_bias']}"
        if self.kwargs['learn_alpha']:
            _repr += f", learn_alpha={self.kwargs['learn_alpha']}"
        else:
            _repr += f", alpha={self.alpha.item():.2f}"
        if self.kwargs['learn_beta']:
            _repr += f", learn_beta={self.kwargs['learn_beta']}"
        else:
            _repr += f", beta={self.beta.item():.2f}"
        _repr += f", activation={self.activation.__class__.__name__}"
        return _repr



from typing import Optional

import torch
from .base import NamedModule, SizedModule


class NamedModuleWrapper(NamedModule):
    """
    Wrapper for a module that does not inherit from NamedModule.
    """

    def __init__(self, module: torch.nn.Module, name: Optional[str] = None):
        super().__init__(name=name)
        self.module = module

    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)


class SizedModuleWrapper(SizedModule):
    """
    Wrapper for a module that does not inherit from SizedModule.
    """

    def __init__(
            self,
            module: torch.nn.Module,
            input_size: Optional[int] = None,
            output_size: Optional[int] = None,
            name: Optional[str] = None
    ):
        super().__init__(input_size=input_size, output_size=output_size, name=name)
        self.module = module

        # TODO:
        # if self.input_size is None:
        # 	self.infer_input_size()

        if self.input_size is not None and self.output_size is None:
            self.infer_output_size()

    def infer_input_size(self):
        raise NotImplementedError("Input size cannot be inferred from a module.")

    def infer_output_size(self):
        if self.input_size is None:
            raise ValueError("Input size is not defined.")
        forward_tensor = torch.zeros((1, int(self.input_size))).to(self.device)
        out = self.module(forward_tensor)
        self.output_size = out[0].shape
        return self.output_size

    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)



from setuptools import setup
# from neurotorch import __author__, __url__, __email__, __version__, __license__
import setuptools
# from src import neurotorch

# with open("README.md", "r", encoding="utf-8") as fh:
#     long_description = fh.read()
#
# with open("requirements.txt", "r", encoding="utf-8") as fh:
#     install_requires = fh.readlines()

setup(
    name='NeuroTorch',
    # version=__version__,
    # description=neurotorch.__doc__,
    long_description='file: README.md',
    long_description_content_type="text/markdown",
    # url=__url__,
    # author=__author__,
    # author_email=__email__,
    # license=__license__,
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
    ],
    # install_requires=install_requires,
    project_urls={
        'Homepage': 'https://github.com/NeuroTorch/NeuroTorch',
        'Source': 'https://github.com/NeuroTorch/NeuroTorch',
        'Documentation': 'https://neurotorch.github.io/NeuroTorch',
    },
)


# build library
#  setup.py sdist bdist_wheel
# With pyproject.toml
# python -m pip install --upgrade build
# python -m build

# publish on PyPI
#   twine check dist/*
#   twine upload --repository-url https://test.pypi.org/legacy/ dist/*
#   twine upload dist/*




from typing import Callable, Dict, List, Optional, Union, Any

import torch

from . import Trainer
from ..metrics import RegressionMetrics


class RegressionTrainer(Trainer):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("predict_method", "get_prediction_trace")
        super().__init__(*args, **kwargs)

    @staticmethod
    def _set_default_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
        kwargs = Trainer._set_default_kwargs(kwargs)
        kwargs.setdefault("foresight_time_steps", None)
        return kwargs

    def _set_default_criterion(self, criterion: Optional[torch.nn.Module]) -> torch.nn.Module:
        if criterion is None:
            if isinstance(self.model.output_sizes, dict):
                criterion = {
                    k: torch.nn.MSELoss() for k in self.model.output_sizes
                }
            elif isinstance(self.model.output_sizes, int):
                criterion = torch.nn.MSELoss()
            else:
                raise ValueError("Unknown criterion type")
        return criterion

    def _set_default_metrics(self, metrics: Optional[List[Callable]]):
        if metrics is None:
            metrics = [RegressionMetrics(self.model)]
        return metrics





import pickle
import time
import warnings
from typing import Callable, Dict, List, Any, Tuple, Union, Iterable, Optional, Sequence

import numpy as np
import torch
import torchvision
from matplotlib import pyplot as plt

from ..transforms.base import to_tensor, to_numpy


def set_seed(seed: int):
    """
    Set the seed of the random number generator.

    :param seed: The seed to set.
    """
    import random
    import torch
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def unitary_rn_normal_matrix(n: int, m: int, generator: Optional[torch.Generator] = None) -> torch.Tensor:
    max_dim, min_dim = max(n, m), min(n, m)

    # rn_matrix = torch.randn((n, m), generator=generator)
    # u, s, v = torch.linalg.svd(rn_matrix, full_matrices=True)
    # unitary_rn_matrix = u @ torch.eye(*rn_matrix.shape) @ v

    # rn_matrix = torch.randn((n, m), generator=generator)
    # u, s, v = torch.linalg.svd(rn_matrix, full_matrices=True)
    # eye = torch.eye(min_dim, max_dim)
    # if n > m:
    # 	unitary_rn_matrix = v @ eye @ u
    # else:
    # 	unitary_rn_matrix = u @ eye @ v

    # rn_matrix = torch.randn((max_dim, max_dim), generator=generator)
    # rn_matrix[:max_dim-min_dim] = 0.0
    # u, s, v = torch.linalg.svd(rn_matrix, full_matrices=False)
    # unitary_rn_matrix = v[:min_dim]

    rn_matrix = torch.randn((n, m), generator=generator)
    u, s, v = torch.linalg.svd(rn_matrix, full_matrices=False)
    unitary_rn_matrix = u @ v

    if tuple(to_numpy(unitary_rn_matrix.shape, dtype=int)) != (n, m):
        unitary_rn_matrix = unitary_rn_matrix.T
    return unitary_rn_matrix


def format_pseudo_rn_seed(seed: Optional[int] = None) -> int:
    """
    Format the pseudo random number generator seed. If the seed is None, return a pseudo random seed
    else return the given seed.

    :param seed: The seed to format.
    :type seed: int or None

    :return: The formatted seed.
    :rtype: int
    """
    import random
    if seed is None:
        seed = int(time.time()) + random.randint(0, np.iinfo(int).max)
    assert isinstance(seed, int), "Seed must be an integer."
    return seed







from typing import Iterable, Optional, Iterator, Dict, Any, List


class BaseCallback:
    """
    Class used to create a callback that can be used to monitor or modify the training process.

    Training Phases:
        - Iteration: One full pass through the training dataset and the validation dataset.
        - Epoch: One full pass through the training dataset or the validation dataset.
        - Batch: One forward pass through the network.
        - Train: One full pass through the training dataset.
        - Validation: One full pass through the validation dataset.

    Callbacks methods are called in the following order:
        - :meth:`start`
        - :meth:`load_checkpoint_state`
        * Executes n_iterations times:
            - :meth:`on_iteration_begin`
            - :meth:`on_train_begin`
            * Executes n_epochs times:
                - :meth:`on_epoch_begin`
                * Executes n_batches times:
                    - :meth:`on_batch_begin`
                    - :meth:`on_optimization_begin`
                    - :meth:`on_optimization_end`
                    - :meth:`on_batch_end`
                - :meth:`on_epoch_end`
            - :meth:`on_train_end`
            - :meth:`on_validation_begin`
            - :meth:`on_epoch_begin`
            * Executes n_batches times:
                - :meth:`on_batch_begin`
                - :meth:`on_validation_batch_begin`
                - :meth:`on_validation_batch_end`
                - :meth:`on_batch_end`
            - :meth:`on_epoch_end`
            - :meth:`on_validation_end`
            - :meth:`on_iteration_end`
        - :meth:`close`

    :Note: The special method :meth:`get_checkpoint_state` is called by the object :class:`CheckpointManager` to
        save the state of the callback in the checkpoint file. The when this method is called is then determined by
        the :class:`CheckpointManager` object if it is used in the trainer callbacks. In the same way, the method
        :meth:`load_checkpoint_state` is called by the :class:`CheckpointManager` to load the state of the callback
        from the checkpoint file if it is used in the trainer callbacks.

    :Note: The special method :meth:`__del__` is called when the callback is deleted. This is used to call the
        :meth:`close` method if it was not called before.

    :Attributes:
        - :attr:`Priority`: The priority of the callback. The lower the priority, the earlier the callback is called.
            Default is 10.
    """
    DEFAULT_PRIORITY = 10
    DEFAULT_LOW_PRIORITY = 100
    DEFAULT_MEDIUM_PRIORITY = 50
    DEFAULT_HIGH_PRIORITY = 0

    instance_counter = 0

    UNPICKEABLE_ATTRIBUTES = ["trainer"]

    def __init__(
            self,
            priority: Optional[int] = None,
            name: Optional[str] = None,
            save_state: bool = True,
            load_state: Optional[bool] = None,
            **kwargs
    ):
        """
        :param priority: The priority of the callback. The lower the priority, the earlier the callback is called.
            At the beginning of the training the priorities of the callbacks are reversed for the :meth:`load_state`
            method. Default is 10.
        :type priority: int, optional
        :param name: The name of the callback. If None, the name is set to the class name. Default is None.
        :type name: str, optional
        :param save_state: If True, the state of the callback is saved in the checkpoint file. Default is True.
        :type save_state: bool, optional
        :param load_state: If True, the state of the callback is loaded from the checkpoint file. Default is equal to
            save_state.
        :type load_state: bool, optional
        """
        self.kwargs = kwargs
        self.priority = priority if priority is not None else self.DEFAULT_PRIORITY
        self.instance_id = self.instance_counter
        self.name = name if name is not None else f"{self.__class__.__name__}<{self.instance_id}>"
        self.save_state = save_state
        self.load_state = load_state if load_state is not None else save_state
        self.__class__.instance_counter += 1
        self.trainer = None
        self._start_flag = False
        self._close_flag = False

    def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
        """
        Loads the state of the callback from a dictionary.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param checkpoint: The dictionary containing all the states of the trainer.
        :type checkpoint: dict

        :return: None
        """
        if self.load_state and checkpoint is not None:
            state = checkpoint.get(self.name, None)
            if state is not None:
                self.__dict__.update(state)

    def get_checkpoint_state(self, trainer, **kwargs) -> object:
        """
        Get the state of the callback. This is called when the checkpoint manager saves the state of the trainer.
        Then this state is saved in the checkpoint file with the name of the callback as the key.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: The state of the callback.
        :rtype: An pickleable object.
        """
        if self.save_state:
            return {k: v for k, v in self.__dict__.items() if k not in self.UNPICKEABLE_ATTRIBUTES}

    def start(self, trainer, **kwargs):
        """
        Called when the training starts. This is the first callback called.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        self.trainer = trainer
        self._start_flag = True

    def close(self, trainer, **kwargs):
        """
        Called when the training ends. This is the last callback called.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        self._close_flag = True

    def on_train_begin(self, trainer, **kwargs):
        """
        Called when the train phase of an iteration starts. The train phase is defined as a full pass through the
        training dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_train_end(self, trainer, **kwargs):
        """
        Called when the train phase of an iteration ends. The train phase is defined as a full pass through the
        training dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_epoch_begin(self, trainer, **kwargs):
        """
        Called when an epoch starts. An epoch is defined as one full pass through the training dataset or
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_epoch_end(self, trainer, **kwargs):
        """
        Called when an epoch ends. An epoch is defined as one full pass through the training dataset or
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_batch_begin(self, trainer, **kwargs):
        """
        Called when a batch starts. The batch is defined as one forward pass through the network.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_batch_end(self, trainer, **kwargs):
        """
        Called when a batch ends. The batch is defined as one forward pass through the network.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_validation_begin(self, trainer, **kwargs):
        """
        Called when the validation phase of an iteration starts. The validation phase is defined as a full pass through
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return:
        """
        pass

    def on_validation_end(self, trainer, **kwargs):
        """
        Called when the validation phase of an iteration ends. The validation phase is defined as a full pass through
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_iteration_begin(self, trainer, **kwargs):
        """
        Called when an iteration starts. An iteration is defined as one full pass through the training dataset and
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_iteration_end(self, trainer, **kwargs):
        """
        Called when an iteration ends. An iteration is defined as one full pass through the training dataset and
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_optimization_begin(self, trainer, **kwargs):
        """
        Called when the optimization phase of an iteration starts. The optimization phase is defined as
        the moment where the model weights are updated.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param kwargs: Additional arguments.

        :keyword x: The input data.
        :keyword y: The target data.
        :keyword pred: The predicted data.

        :return: None
        """
        pass

    def on_optimization_end(self, trainer, **kwargs):
        """
        Called when the optimization phase of an iteration ends. The optimization phase is defined as
        the moment where the model weights are updated.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_validation_batch_begin(self, trainer, **kwargs):
        """
        Called when the validation batch starts. The validation batch is defined as one forward pass through the network
        on the validation dataset. This is used to update the batch loss and metrics on the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param kwargs: Additional arguments.

        :keyword x: The input data.
        :keyword y: The target data.
        :keyword pred: The predicted data.

        :return: None
        """
        pass

    def on_validation_batch_end(self, trainer, **kwargs):
        """
        Called when the validation batch ends. The validation batch is defined as one forward pass through the network
        on the validation dataset. This is used to update the batch loss and metrics on the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        pass

    def on_trajectory_end(self, trainer, trajectory, **kwargs) -> List[Dict[str, Any]]:
        """
        Called when a trajectory ends. This is used in reinforcement learning to update the trajectory loss and metrics.
        Must return a list of dictionaries containing the trajectory metrics. The list must have the same length as the
        trajectory. Each item in the list will update the attribute `others` of the corresponding Experience.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param trajectory: The trajectory i.e. the sequence of Experiences.
        :type trajectory: Trajectory
        :param kwargs: Additional arguments.

        :return: A list of dictionaries containing the trajectory metrics.
        """
        pass

    def on_pbar_update(self, trainer, **kwargs) -> dict:
        """
        Called when the progress bar is updated.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param kwargs: Additional arguments.

        :return: None
        """
        return {}

    def __del__(self):
        try:
            if (not self._close_flag) and self.trainer is not None:
                self.close(self.trainer)
        except:
            pass
        self.__class__.instance_counter -= 1

    def extra_repr(self) -> str:
        return ""

    def __repr__(self):
        repr_str = f"{self.name}: ("
        repr_str += f"priority={self.priority}, "
        repr_str += f"save_state={self.save_state}, "
        repr_str += f"load_state={self.load_state}, "
        repr_str += self.extra_repr()
        repr_str += f")"
        return repr_str


class CallbacksList:
    """
    This class is used to store the callbacks that are used during the training. Each callback of the list is called
    in the order they are stored in the list.

    :Attributes:
        - :attr:`callbacks` (List[BaseCallback]): The callbacks to use.
    """
    def __init__(self, callbacks: Optional[Iterable[BaseCallback]] = None):
        """
        Constructor of the CallbacksList class.

        :param callbacks: The callbacks to use.
        :type callbacks: Iterable[BaseCallback]
        """
        if callbacks is None:
            callbacks = []
        assert isinstance(callbacks, Iterable), "callbacks must be an Iterable"
        assert all(isinstance(callback, BaseCallback) for callback in callbacks), \
            "All callbacks must be instances of BaseCallback"
        self.callbacks = list(callbacks)
        self._length = len(self.callbacks)
        self.sort_callbacks_()

    def sort_callbacks_(self, reverse: bool = False) -> 'CallbacksList':
        """
        Sorts the callbacks by their priority.

        :param reverse: If True, the callbacks are sorted in descending order.
        :type reverse: bool

        :return: self
        :rtype: CallbacksList
        """
        self.callbacks.sort(key=lambda callback: callback.priority, reverse=reverse)
        return self

    def __getitem__(self, item: int) -> BaseCallback:
        """
        Get a callback from the list.

        :param item: The index of the callback to get.
        :type item: int

        :return: The callback at the given index.
        :rtype: BaseCallback
        """
        return self.callbacks[item]

    def __iter__(self) -> Iterator[BaseCallback]:
        """
        Get an iterator over the callbacks.

        :return: An iterator over the callbacks.
        :rtype: Iterator[BaseCallback]
        """
        return iter(self.callbacks)

    def __len__(self) -> int:
        """
        Get the number of callbacks in the list.

        :return: The number of callbacks in the list.
        :rtype: int
        """
        return self._length

    def __repr__(self):
        repr_str = f"{self.__class__.__name__}(\n"
        for callback in self.callbacks:
            repr_str += f"\t{repr(callback)}, \n"
        repr_str += f")"
        return repr_str

    def append(self, callback: BaseCallback):
        """
        Append a callback to the list.

        :param callback: The callback to append.
        :type callback: BaseCallback

        :return: None
        """
        assert isinstance(callback, BaseCallback), "callback must be an instance of BaseCallback"
        self.callbacks.append(callback)
        self._length += 1
        self.sort_callbacks_()

    def remove(self, callback: BaseCallback):
        """
        Remove a callback from the list.

        :param callback: The callback to remove.
        :type callback: BaseCallback

        :return: None
        """
        assert isinstance(callback, BaseCallback), "callback must be an instance of BaseCallback"
        self.callbacks.remove(callback)
        self._length -= 1
        self.sort_callbacks_()

    def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
        """
        Loads the state of the callback from a dictionary.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param checkpoint: The dictionary containing all the states of the trainer.
        :type checkpoint: dict

        :return: None
        """
        for callback in self.callbacks:
            callback.load_checkpoint_state(trainer, checkpoint, **kwargs)

    def get_checkpoint_state(self, trainer, **kwargs) -> Dict[str, Any]:
        """
        Collates the states of the callbacks. This is called when the checkpoint manager saves the state of the trainer.
        Then those states are saved in the checkpoint file with the name of the callback as the key.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: The state of the callback.
        :rtype: An pickleable dict.
        """
        states = {
            callback.name: callback.get_checkpoint_state(trainer, **kwargs)
            for callback in self.callbacks
        }
        states = {key: value for key, value in states.items() if value is not None}
        return states

    def start(self, trainer, **kwargs):
        """
        Called when the trainer starts.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.start(trainer, **kwargs)

    def close(self, trainer, **kwargs):
        """
        Called when the trainer closes.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.close(trainer, **kwargs)

    def on_train_begin(self, trainer, **kwargs):
        """
        Called when the train phase of an iteration starts. The train phase is defined as a full pass through the
        training dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_train_begin(trainer, **kwargs)

    def on_train_end(self, trainer, **kwargs):
        """
        Called when the train phase of an iteration ends. The train phase is defined as a full pass through the
        training dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_train_end(trainer, **kwargs)

    def on_epoch_begin(self, trainer, **kwargs):
        """
        Called when an epoch starts. An epoch is defined as one full pass through the training dataset or
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_epoch_begin(trainer, **kwargs)

    def on_epoch_end(self, trainer, **kwargs):
        """
        Called when an epoch ends. An epoch is defined as one full pass through the training dataset or
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_epoch_end(trainer, **kwargs)

    def on_batch_begin(self, trainer, **kwargs):
        """
        Called when a batch starts. The batch is defined as one forward pass through the network.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_batch_begin(trainer, **kwargs)

    def on_batch_end(self, trainer, **kwargs):
        """
        Called when a batch ends. The batch is defined as one forward pass through the network.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_batch_end(trainer, **kwargs)

    def on_validation_begin(self, trainer, **kwargs):
        """
        Called when the validation phase of an iteration starts. The validation phase is defined as a full pass through
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_validation_begin(trainer, **kwargs)

    def on_validation_end(self, trainer, **kwargs):
        """
        Called when the validation phase of an iteration ends. The validation phase is defined as a full pass through
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_validation_end(trainer, **kwargs)

    def on_iteration_begin(self, trainer, **kwargs):
        """
        Called when an iteration starts. An iteration is defined as one full pass through the training dataset and
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_iteration_begin(trainer, **kwargs)

    def on_iteration_end(self, trainer, **kwargs):
        """
        Called when an iteration ends. An iteration is defined as one full pass through the training dataset and
        the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_iteration_end(trainer, **kwargs)

    def on_optimization_begin(self, trainer, **kwargs):
        """
        Called when the optimization phase of an iteration starts. The optimization phase is defined as
        the moment where the model weights are updated.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param kwargs: Additional arguments.

        :return: None
        """
        for callback in self.callbacks:
            callback.on_optimization_begin(trainer, **kwargs)

    def on_optimization_end(self, trainer, **kwargs):
        """
        Called when the optimization phase of an iteration ends. The optimization phase is defined as
        the moment where the model weights are updated.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_optimization_end(trainer, **kwargs)

    def on_validation_batch_begin(self, trainer, **kwargs):
        """
        Called when the validation batch starts. The validation batch is defined as one forward pass through the network
        on the validation dataset. This is used to update the batch loss and metrics on the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param kwargs: Additional arguments.

        :keyword x: The input data.
        :keyword y: The target data.
        :keyword pred: The predicted data.

        :return: None
        """
        for callback in self.callbacks:
            callback.on_validation_batch_begin(trainer, **kwargs)

    def on_validation_batch_end(self, trainer, **kwargs):
        """
        Called when the validation batch ends. The validation batch is defined as one forward pass through the network
        on the validation dataset. This is used to update the batch loss and metrics on the validation dataset.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        for callback in self.callbacks:
            callback.on_validation_batch_end(trainer, **kwargs)

    def on_trajectory_end(self, trainer, trajectory, **kwargs) -> List[Dict[str, Any]]:
        """
        Called when a trajectory ends. This is used in reinforcement learning to update the trajectory loss and metrics.
        Must return a  list of dictionaries containing the trajectory metrics. The list must have the same
        length as the trajectory. Each item in the list will update the attribute `others` of the corresponding
        Experience.

        :Note: If the callbacks return the same keys in the dictionaries, the values will be updated, so the
            last callback will prevail.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param trajectory: The trajectory i.e. the sequence of Experiences.
        :type trajectory: Trajectory
        :param kwargs: Additional arguments.

        :return: A list of dictionaries containing the trajectory metrics.
        """
        # re_list = [{}] * len(trajectory)
        re_list = [{} for _ in range(len(trajectory))]
        for callback in self.callbacks:
            callback_return = callback.on_trajectory_end(trainer, trajectory, **kwargs)
            if callback_return is not None:
                assert len(callback_return) == len(trajectory), \
                    "The callback must return a list of dictionaries with the same length as the trajectory."
                for i, re in enumerate(callback_return):
                    re_list[i].update(re)
        return re_list

    def on_pbar_update(self, trainer, **kwargs) -> dict:
        """
        Called when the progress bar is updated.

        :param trainer: The trainer.
        :type trainer: Trainer
        :param kwargs: Additional arguments.

        :return: None
        """
        re_dict = {}
        for callback in self.callbacks:
            callback_dict = callback.on_pbar_update(trainer, **kwargs)
            if callback_dict is None:
                callback_dict = {}
            elif not isinstance(callback_dict, dict):
                callback_dict = {callback.name: callback_dict}
            re_dict.update(callback_dict)
        return re_dict




from typing import Callable

import torch


class CallableToModuleWrapper(torch.nn.Module):
    def __init__(self, callable_object: Callable):
        super().__init__()
        assert callable(callable_object), "The callable object must be callable."
        self.callable_object = callable_object

    def __repr__(self):
        return f"CallableToModuleWrapper({self.callable_object})"

    def forward(self, *args, **kwargs):
        return self.callable_object(*args, **kwargs)


